# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from requests.models import Response
from unittest import mock

from planet_auth.oidc.api_clients.api_client import OidcApiClient, OidcApiClientException

TEST_API_ENDPOINT = "https://blackhole.unittest.planet.com/api"
TEST_DATA1 = {"key1": "value1"}


def mocked_requests_get_or_post(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    # request_url = kwargs.get('url', None)
    request_params = kwargs.get("params", None)
    request_data = kwargs.get("data", None)

    if request_url == TEST_API_ENDPOINT + "/data1.json":
        response.headers["content-type"] = "application/json"
        response._content = str.encode(json.dumps(TEST_DATA1))
    elif request_url == TEST_API_ENDPOINT + "/data2.dat":
        response.headers["content-type"] = "data/test"
        response._content = b"1234567890"
    elif request_url == TEST_API_ENDPOINT + "/reflect1.json":
        # for reflecting from params
        response._content = str.encode(json.dumps(request_params.get("reflect_data")))
        response.status_code = request_params.get("status_code")
    elif request_url == TEST_API_ENDPOINT + "/reflect2.json":
        # for reflecting from data
        response._content = str.encode(json.dumps(request_data.get("reflect_data")))
        response.status_code = request_data.get("status_code")
    elif request_url == TEST_API_ENDPOINT + "/oidc_error_1.json":
        response.headers["content-type"] = "application/json"
        response._content = str.encode(
            json.dumps({"error": "TEST_ERROR", "error_description": "This is a test error"})
        )
    elif request_url == TEST_API_ENDPOINT + "/oidc_error_2.json":
        response.headers["content-type"] = "application/json"
        response._content = str.encode(json.dumps({"errorCode": "TEST_ERROR", "errorSummary": "This is a test error"}))
    elif request_url == TEST_API_ENDPOINT + "/empty_json.json":
        response.headers["content-type"] = "application/json"
        response._content = None
    else:
        response.status_code = 404
    return response


class ApiClientBaseTest(unittest.TestCase):
    @mock.patch("requests.sessions.Session.get", side_effect=mocked_requests_get_or_post)
    def test_checked_get_valid(self, mock_get):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/reflect1.json")
        under_test._checked_get(params={"status_code": 200, "reflect_data": "test_data"}, request_auth=None)

    @mock.patch("requests.sessions.Session.get", side_effect=mocked_requests_get_or_post)
    def test_checked_get_invalid(self, mock_get):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/reflect1.json")
        with self.assertRaises(OidcApiClientException):
            under_test._checked_get(params={"status_code": 401, "reflect_data": "test_data"}, request_auth=None)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_requests_get_or_post)
    def test_checked_post_valid(self, mock_get):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/reflect2.json")
        under_test._checked_post(params={"status_code": 200, "reflect_data": "test_data"}, request_auth=None)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_requests_get_or_post)
    def test_checked_post_invalid(self, mock_get):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/reflect2.json")
        with self.assertRaises(OidcApiClientException):
            under_test._checked_post(params={"status_code": 401, "reflect_data": "test_data"}, request_auth=None)

    @mock.patch("requests.sessions.Session.get", side_effect=mocked_requests_get_or_post)
    def test_checked_get_json_response_valid(self, mock_get):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/data1.json")
        json_response = under_test._checked_get_json_response(params=None, request_auth=None)
        self.assertEqual(TEST_DATA1, json_response)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_requests_get_or_post)
    def test_checked_post_json_response_valid(self, mock_get):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/data1.json")
        json_response = under_test._checked_post_json_response(params=None, request_auth=None)
        self.assertEqual(TEST_DATA1, json_response)

    @mock.patch("requests.sessions.Session.get", side_effect=mocked_requests_get_or_post)
    def test_checked_get_json_response_invalid(self, mock_get):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/data2.dat")
        with self.assertRaises(OidcApiClientException):
            under_test._checked_get_json_response(params=None, request_auth=None)

    @mock.patch("requests.sessions.Session.post", side_effect=mocked_requests_get_or_post)
    def test_checked_post_json_response_invalid(self, mock_get):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/data2.dat")
        with self.assertRaises(OidcApiClientException):
            under_test._checked_post_json_response(params=None, request_auth=None)

    @mock.patch("requests.sessions.Session.get", side_effect=mocked_requests_get_or_post)
    def test_checked_get_json_response_invalid_empty(self, mock_get):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/empty_json.json")
        with self.assertRaises(OidcApiClientException):
            under_test._checked_get_json_response(params=None, request_auth=None)

    @mock.patch("requests.sessions.Session.get", side_effect=mocked_requests_get_or_post)
    def test_checked_get_json_response_invalid_not_found(self, mock_get):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/not_found")
        with self.assertRaises(OidcApiClientException):
            under_test._checked_get_json_response(params=None, request_auth=None)

    @mock.patch("requests.sessions.Session.get", side_effect=mocked_requests_get_or_post)
    def test_check_oidc_error_1(self, get_mock):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/oidc_error_1.json")
        with self.assertRaises(OidcApiClientException):
            under_test._checked_get_json_response(params=None, request_auth=None)

    @mock.patch("requests.sessions.Session.get", side_effect=mocked_requests_get_or_post)
    def test_check_oidc_error_2(self, get_mock):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/oidc_error_2.json")
        with self.assertRaises(OidcApiClientException):
            under_test._checked_get_json_response(params=None, request_auth=None)

    @mock.patch("requests.sessions.Session.get", side_effect=mocked_requests_get_or_post)
    def test_check_oidc_error_content_type(self, get_mock):
        under_test = OidcApiClient(endpoint_uri=TEST_API_ENDPOINT + "/data2.dat")
        with self.assertRaises(OidcApiClientException):
            under_test._checked_get_json_response(params=None, request_auth=None)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC

from planet_auth.oidc.api_clients.api_client import EnricherPayloadType, EnricherReturnType
from planet_auth.auth_client import AuthClientConfigException
from planet_auth.oidc.api_clients.oidc_request_auth import (
    prepare_client_secret_request_auth,
    prepare_client_secret_auth_payload,
)
from planet_auth.oidc.auth_client import OidcAuthClient, OidcAuthClientConfig


# TODO: Can I implement these as RequestAuthenticator? Good dog food.
class OidcAuthClientWithClientSecretClientConfig(OidcAuthClientConfig, ABC):
    """
    Mix-in base class for OIDC auth clients that make use of a client secret.
    """

    def __init__(self, client_secret: str = None, **kwargs):
        super().__init__(**kwargs)
        if client_secret:
            self._data["client_secret"] = client_secret

    def client_secret(self) -> str:
        return self.lazy_get("client_secret")

    def check_data(self, data):
        super().check_data(data)
        if not data.get("client_secret"):
            raise AuthClientConfigException(
                message="client_secret must be configured for {} client.".format(self.__class__.__name__)
            )

    @classmethod
    def meta(cls):
        return {
            "config_hints": super().meta().get("config_hints")
            + [
                {
                    "config_key": "client_secret",
                    "config_key_name": "Client secret",
                    "config_key_description": "Client shared secret known only to the client and the authorization server",
                }
            ]
        }


class OidcAuthClientWithClientSecret_HttpBasicAuthEnrichment(OidcAuthClient, ABC):
    """
    Mix-in base class for OIDC auth clients that make use of a client secret
    to perform client authentication.
    """

    def __init__(self, client_config: OidcAuthClientWithClientSecretClientConfig):
        super().__init__(client_config)
        self._oidc_client_secret_client_config = client_config

    def _client_auth_enricher(self, raw_payload: EnricherPayloadType, audience: str) -> EnricherReturnType:
        return raw_payload, prepare_client_secret_request_auth(
            self._oidc_client_secret_client_config.client_id(), self._oidc_client_secret_client_config.client_secret()
        )


class OidcAuthClientWithClientSecret_PayloadAuthEnrichment(OidcAuthClient, ABC):
    """
    Mix-in base class for OIDC auth clients that make use of a client secret
    to perform client authentication.
    """

    def __init__(self, client_config: OidcAuthClientWithClientSecretClientConfig):
        super().__init__(client_config)
        self._oidc_client_secret_client_config = client_config

    def _client_auth_enricher(self, raw_payload: EnricherPayloadType, audience: str) -> EnricherReturnType:
        auth_payload = prepare_client_secret_auth_payload(
            client_id=self._oidc_client_secret_client_config.client_id(),
            client_secret=self._oidc_client_secret_client_config.client_secret(),
        )
        enriched_payload = {**raw_payload, **auth_payload}
        return enriched_payload, None



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import functools
import json
from typing import Optional

import planet_auth
from planet_auth.constants import AUTH_CONFIG_FILE_SOPS, AUTH_CONFIG_FILE_PLAIN
from planet_auth.util import custom_json_class_dumper

from planet_auth_utils.builtins import Builtins
from planet_auth_utils.profile import Profile
from .prompts import prompt_and_change_user_default_profile_if_different


def recast_exceptions_to_click(*exceptions, **params):  # pylint: disable=W0613
    if not exceptions:
        exceptions = (Exception,)
    # params.get('some_arg', 'default')

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exceptions as e:
                raise click.ClickException(str(e))

        return wrapper

    return decorator


def print_obj(obj):
    json_str = json.dumps(obj, indent=2, sort_keys=True, default=custom_json_class_dumper)
    print(json_str)


def post_login_cmd_helper(
    override_auth_context: planet_auth.Auth, use_sops, prompt_pre_selection: Optional[bool] = None
):
    override_profile_name = override_auth_context.profile_name()
    if not override_profile_name:
        # Can't save to a profile if there is none.  We don't really expect this in the cases
        # where this is used for the CLI, but this keeps linters happy.
        return

    # If someone performed a login with a non-default profile, it's
    # reasonable to ask if they intend to change their defaults.
    prompt_and_change_user_default_profile_if_different(
        candidate_profile_name=override_profile_name, change_default_selection=prompt_pre_selection
    )

    # If the config was created ad-hoc by the factory, the factory does
    # not associate it with a file to support factory use in a context
    # where in-memory operations are desired.  This util function is for
    # helping CLI commands, we can be more opinionated about what is to
    # be done.

    # Don't clobber built-in profiles.
    if not Builtins.is_builtin_profile(override_profile_name):
        if use_sops:
            new_profile_config_file_path = Profile.get_profile_file_path(
                profile=override_profile_name, filename=AUTH_CONFIG_FILE_SOPS
            )
        else:
            new_profile_config_file_path = Profile.get_profile_file_path(
                profile=override_profile_name, filename=AUTH_CONFIG_FILE_PLAIN
            )

        # TODO? should we update if it exists with any command line options?
        #       The way things are currently structured, options added to the
        #       login command (like --scope) are not pushed down into the profile,
        #       but considered runtime overrides.
        if not new_profile_config_file_path.exists():
            override_auth_context.auth_client().config().set_path(new_profile_config_file_path)
            # No need to specify provider. The CLI is only concerned with the
            # default file based storage provider for now.
            # override_auth_context.auth_client().config().set_storage_provider()
            override_auth_context.auth_client().config().save()

    # TODO? Set ctx.obj["AUTH"] to override_auth_context?
    #       Only if they responded "yes" to changing the default?
    #       It doesn't really matter, since we expect the program to exit.



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

import planet_auth.logging.auth_logger

auth_logger = planet_auth.logging.auth_logger.getAuthLogger()


def parse_content_type(content_type: Optional[str]) -> Dict[str, Optional[str]]:
    result: Dict[str, Optional[str]] = {
        "content-type": None,
    }
    if content_type:
        ct = content_type.split(";")
        result["content-type"] = ct.pop(0).strip()
        if not result["content-type"]:
            # Don't return blank strings
            result["content-type"] = None
        for subfield in ct:
            sf = subfield.split("=", 1)
            if sf[0].strip():
                if len(sf) == 1:
                    result[sf[0].strip()] = None
                else:
                    result[sf[0].strip()] = sf[1].strip()
    return result


def custom_json_class_dumper(obj):
    try:
        return obj.__json_pretty_dumps__()
    except Exception:
        return obj



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import pathlib

from planet_auth import constants
from planet_auth_utils.constants import EnvironmentVariables


def opt_api_key(function):
    """
    Click option for specifying an API key
    """
    function = click.option(
        "--auth-api-key",
        type=str,
        envvar=EnvironmentVariables.AUTH_API_KEY,
        help="Specify an API key.",
        default=None,
        show_envvar=True,
        show_default=True,
    )(function)
    return function


def opt_client_id(function):
    """
    Click option for specifying an OAuth client ID.
    """
    function = click.option(
        "--auth-client-id",
        type=str,
        envvar=EnvironmentVariables.AUTH_CLIENT_ID,
        help="Specify the OAuth client ID.",
        default=None,
        show_envvar=True,
        show_default=True,
    )(function)
    return function


def opt_client_secret(function):
    """
    Click option for specifying an OAuth client secret.
    """
    function = click.option(
        "--auth-client-secret",
        type=str,
        envvar=EnvironmentVariables.AUTH_CLIENT_SECRET,
        help="Specify the OAuth client Secret.",
        default=None,
        show_envvar=True,
        show_default=True,
    )(function)
    return function


def opt_profile(function):
    """
    Click option for specifying an auth profile for the
    planet_auth package's click commands.
    """
    function = click.option(
        "--auth-profile",
        type=str,
        envvar=EnvironmentVariables.AUTH_PROFILE,
        help="Select the client profile to use.  User created profiles are "
        f" defined by creating a subdirectory ~/{constants.PROFILE_DIR}/.  Additionally, a number of"
        ' built-in profiles are understood.  See the "profile list" command'
        " for defined profiles.  The auth profile controls how the software"
        " interacts with authentication services, as well as how it"
        " authenticates to other APIs.  If this option is not set,"
        " a profile will be selected according to environment variables or"
        ' a preference registered with the "profile set" command.',
        default=None,
        show_envvar=True,
        show_default=True,
        is_eager=True,
    )(function)
    return function


def opt_organization(function):
    """
    Click option for specifying an Organization.
    """
    function = click.option(
        "--organization",
        multiple=False,
        type=str,
        envvar=EnvironmentVariables.AUTH_ORGANIZATION,
        help="Organization to use when performing authentication.  When present, this option will be"
        " appended to authorization requests.  Not all implementations understand this option.",
        default=None,
        show_envvar=True,
        show_default=True,
    )(function)
    return function


def opt_project(function):
    """
    Click option for specifying a project ID.
    """
    function = click.option(
        "--project",
        multiple=False,
        type=str,
        # envvar=EnvironmentVariables.AUTH_PROJECT,
        help="Project ID to use when performing authentication.  When present, this option will be"
        " appended to authorization requests.  Not all implementations understand this option.",
        default=None,
        show_envvar=True,
        show_default=True,
    )(function)
    return function


# TODO -  Consider switching to click prompts where we current rely on the lower level planet_auth
#         to prompt the user. Currently, some of this IO is delegated to the planet_auth library.
#         I generally think user IO belongs with the app, and not the the library, but since the
#         lib also handles things like browser interaction this is not entirely easy to abstract
#         away.
def opt_password(hidden=True):
    def decorator(function):
        """
        Click option for specifying a password for the
        planet_auth package's click commands.
        """
        function = click.option(
            "--password",
            type=str,
            envvar=EnvironmentVariables.AUTH_PASSWORD,
            help="Password used for authentication.  May not be used by all authentication mechanisms.",
            default=None,
            show_envvar=True,
            show_default=True,
            hidden=hidden,  # Primarily used by legacy auth.  OAuth2 is preferred, wherein we do not handle username/password.
        )(function)
        return function

    return decorator


def opt_username(hidden=True):
    def decorator(function):
        """
        Click option for specifying a username for the
        planet_auth package's click commands.
        """
        function = click.option(
            "--username",
            "--email",
            type=str,
            envvar=EnvironmentVariables.AUTH_USERNAME,
            help="Username used for authentication.  May not be used by all authentication mechanisms.",
            default=None,
            show_envvar=True,
            show_default=True,
            hidden=hidden,  # Primarily used by legacy auth.  OAuth2 is preferred, wherein we do not handle username/password.
        )(function)
        return function

    return decorator


def opt_loglevel(function):
    """
    Click option for specifying a log level.
    """
    function = click.option(
        "-l",
        "--loglevel",
        envvar=EnvironmentVariables.AUTH_LOGLEVEL,
        help="Set the log level.",
        type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"], case_sensitive=False),
        default="INFO",
        show_envvar=True,
        show_default=True,
    )(function)
    return function


def opt_yes_no(function):
    """
    Click option to bypass prompts with a yes or no selection.
    """
    function = click.option(
        "--yes/--no",
        "-y/-n",
        help='Skip user prompts with a "yes" or "no" selection',
        default=None,
        show_default=True,
    )(function)
    return function


def opt_human_readable(function):
    """
    Click option to toggle raw / human-readable formatting.
    """
    function = click.option(
        "--human-readable/--no-human-readable",
        "-H",
        help="Reformat fields to be human readable.",
        default=False,
        show_default=True,
    )(function)
    return function


def opt_long(function):
    """
    Click option specifying that long or more detailed output should be produced.
    """
    function = click.option(
        "-l",
        "--long",
        help="Longer, more detailed output",
        is_flag=True,
        default=False,
        show_default=True,
    )(function)
    return function


def opt_open_browser(function):
    """
    Click option for specifying whether or not opening a browser is permitted
    for the planet_auth package's click commands.
    """
    function = click.option(
        "--open-browser/--no-open-browser",
        help="Allow/Suppress the automatic opening of a browser window.",
        default=True,
        show_default=True,
    )(function)
    return function


def opt_show_qr_code(function):
    """
    Click option for specifying whether or not a QR code should be displayed.
    """
    function = click.option(
        "--show-qr-code/--no-show-qr-code",
        help="Control whether a QR code is displayed for the user.",
        default=False,
        show_default=True,
    )(function)
    return function


def opt_token(function):
    """
    Click option for specifying a token literal.
    """
    function = click.option(
        "--token",
        help="Token string.",
        type=str,
        # envvar=EnvironmentVariables.AUTH_TOKEN,
        show_envvar=False,
        show_default=False,
    )(function)
    return function


def opt_token_file(function):
    """
    Click option for specifying a token file location for the
    planet_auth package's click commands.
    """
    function = click.option(
        "--token-file",
        type=click.Path(exists=True, file_okay=True, readable=True, path_type=pathlib.Path),
        envvar=EnvironmentVariables.AUTH_TOKEN_FILE,
        help="File containing a token.",
        default=None,
        show_envvar=False,
        show_default=True,
    )(function)
    return function


def opt_issuer(required=False):
    def decorator(function):
        """
        Click option for specifying an OAuth token issuer for the
        planet_auth package's click commands.
        """
        function = click.option(
            "--issuer",
            type=str,
            envvar=EnvironmentVariables.AUTH_ISSUER,
            help="Token issuer.",
            default=None,
            show_envvar=False,
            show_default=False,
            required=required,
        )(function)
        return function

    return decorator


def opt_audience(required=False):
    def decorator(function):
        """
        Click option for specifying an OAuth token audience for the
        planet_auth package's click commands.
        """
        function = click.option(
            "--audience",
            multiple=True,
            type=str,
            envvar=EnvironmentVariables.AUTH_AUDIENCE,
            help="Token audiences.  Specify multiple options to set"
            " multiple audiences.  When set via environment variable, audiences"
            " should be white space delimited.",
            default=None,
            show_envvar=True,
            show_default=True,
            required=required,
        )(function)
        return function

    return decorator


def opt_refresh(function):
    """
    Click option specifying a refresh should be attempted if applicable.
    """
    function = click.option(
        "--refresh/--no-refresh",
        help="Automatically perform a credential refresh if required.",
        default=True,
        show_default=True,
    )(function)
    return function


def opt_scope(function):
    """
    Click option for specifying an OAuth token scope for the
    planet_auth package's click commands.
    """
    function = click.option(
        "--scope",
        multiple=True,
        type=str,
        envvar=EnvironmentVariables.AUTH_SCOPE,
        help="Token scope.  Specify multiple options to specify"
        " multiple scopes.  When set via environment variable, scopes"
        " should be white space delimited.  Default value is determined"
        " by the selected auth profile.",
        default=None,
        show_envvar=True,
        show_default=True,
    )(function)
    return function


def opt_sops(function):
    """
    Click option specifying that SOPS should be used.
    """
    function = click.option(
        "--sops/--no-sops",
        help="Use sops when creating new files where applicable."
        " The environment must be configured for SOPS to work by default.",
        default=False,
        show_default=True,
    )(function)
    return function


def opt_extra(function):
    """
    Click option for specifying extra options.
    """
    function = click.option(
        "--extra",
        "-O",
        multiple=True,
        type=str,
        envvar=EnvironmentVariables.AUTH_EXTRA,
        help="Specify an extra option.  Specify multiple options to specify"
        " multiple extra options.  The format of an option is <key>=<value>."
        " When set via environment variable, values should be delimited by"
        " whitespace.",
        default=None,
        show_envvar=True,
        show_default=True,
    )(function)
    return function



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import shutil
import unittest

from planet_auth.auth_client import AuthClientException
from planet_auth.constants import AUTH_CONFIG_FILE_PLAIN
from planet_auth.static_api_key.auth_client import StaticApiKeyAuthClientConfig

from planet_auth_utils.profile import Profile, ProfileException
from tests.test_planet_auth_utils.util import tdata_resource_file_path, TestWithHomeDirProfiles


PROFILE1_NAME = "profile1_static_api_key"
PROFILE2_NAME = "profile2_invalid"
PROFILE3_NAME = "profile3_empty"


class ProfileTest(TestWithHomeDirProfiles, unittest.TestCase):
    def setUp(self):
        self.setUp_testHomeDir()

        self.profile1_dir_path = self.mkProfileDir(PROFILE1_NAME)
        shutil.copy(
            tdata_resource_file_path("auth_client_configs/utest/static_api_key.json"),
            self.profile1_dir_path.joinpath(AUTH_CONFIG_FILE_PLAIN),
        )

        self.profile2_dir_path = self.mkProfileDir(PROFILE2_NAME)
        shutil.copy(
            tdata_resource_file_path("auth_client_configs/utest/invalid_client_config.json"),
            self.profile2_dir_path.joinpath(AUTH_CONFIG_FILE_PLAIN),
        )

        self.profile3_dir_path = self.mkProfileDir(PROFILE3_NAME)

    def tearDown(self) -> None:
        self.tearDown_testHomeDir()

    def test_filepath_no_profile_no_override(self):
        with self.assertRaises(ProfileException):
            Profile.get_profile_file_path(filename="testfile.dat", profile=None, override_path=None)
        # Old behavior
        # self.assertIsInstance(under_test, pathlib.Path)
        # self.assertEqual(pathlib.Path.home().joinpath(".planet/default/testfile.dat"), under_test)

    def test_filepath_blank_profile_no_override(self):
        with self.assertRaises(ProfileException):
            Profile.get_profile_file_path(filename="testfile.dat", profile="", override_path=None)
        # Old behavior
        # self.assertIsInstance(under_test, pathlib.Path)
        # self.assertEqual(pathlib.Path.home().joinpath(".planet/default/testfile.dat"), under_test)

    def test_filepath_blank_profile_with_override(self):
        under_test = Profile.get_profile_file_path(filename="testfile.dat", profile="", override_path="/override")
        self.assertIsInstance(under_test, pathlib.Path)
        self.assertEqual(pathlib.Path("/override"), under_test)

    def test_filepath(self):
        under_test = Profile.get_profile_file_path(filename="testfile.dat", profile="test_profile", override_path=None)
        self.assertIsInstance(under_test, pathlib.Path)
        self.assertEqual(pathlib.Path(".planet/test_profile/testfile.dat"), under_test)

    def test_pathfile_override(self):
        under_test = Profile.get_profile_file_path(
            filename="testfile.dat", profile="test_profile", override_path="/override"
        )
        self.assertIsInstance(under_test, pathlib.Path)
        self.assertEqual(pathlib.Path("/override"), under_test)

    def test_priority_path_with_override(self):
        under_test = Profile.get_profile_file_path_with_priority(
            filenames=["does_not_exist_1", "does_not_exist_2"], profile="test_profile", override_path="/override"
        )
        self.assertEqual(pathlib.Path("/override"), under_test)

    def test_priority_path_fallback_behavior(self):
        under_test = Profile.get_profile_file_path_with_priority(
            filenames=["does_not_exist_1", "does_not_exist_2"], profile="test_profile", override_path=None
        )
        self.assertEqual(pathlib.Path(".planet/test_profile/does_not_exist_2"), under_test)

    def test_priority_path_first_choice_wins_if_exists(self):
        profile_dir = pathlib.Path.home().joinpath(".planet/test_profile")
        profile_dir.mkdir(parents=True, exist_ok=True)
        pathlib.Path.home().joinpath(".planet/test_profile/does_exist_1").touch()
        under_test = Profile.get_profile_file_path_with_priority(
            filenames=["does_exist_1", "does_not_exist_2"], profile="test_profile", override_path=None
        )
        self.assertEqual(pathlib.Path(".planet/test_profile/does_exist_1"), under_test)

    def test_load_client_none(self):
        with self.assertRaises(ProfileException):
            Profile.load_auth_client_config(profile=None)

    def test_load_client_custom_profile_ok(self):
        client_conf = Profile.load_auth_client_config(profile=PROFILE1_NAME)
        self.assertIsInstance(client_conf, StaticApiKeyAuthClientConfig)

    def test_load_client_custom_profile_invalid(self):
        with self.assertRaises(AuthClientException):
            Profile.load_auth_client_config(profile=PROFILE2_NAME)

    def test_load_client_custom_profile_does_not_exist(self):
        with self.assertRaises(FileNotFoundError):
            Profile.load_auth_client_config(profile="profile_does_not_exist")

    def test_list_profiles(self):
        profile_list = Profile.list_on_disk_profiles()
        self.assertEqual(profile_list, [PROFILE1_NAME, PROFILE2_NAME])



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import click
import sys

from planet_auth import (
    AuthException,
    FileBackedPlanetLegacyApiKey,
    PlanetLegacyAuthClient,
    PlanetLegacyAuthClientConfig,
    StaticApiKeyAuthClient,
    StaticApiKeyAuthClientConfig,
)

from .options import opt_password, opt_sops, opt_username, opt_yes_no
from .util import recast_exceptions_to_click, post_login_cmd_helper


def _check_client_type_pllegacy(ctx):
    if not isinstance(ctx.obj["AUTH"].auth_client(), PlanetLegacyAuthClient):
        raise click.ClickException(
            f'"legacy" auth commands can only be used with "{PlanetLegacyAuthClientConfig.meta()["client_type"]}" type auth profiles.'
            f' The current profile "{ctx.obj["AUTH"].profile_name()}" is of type "{ctx.obj["AUTH"].auth_client()._auth_client_config.meta()["client_type"]}".'
        )


def _check_client_type_pllegacy_or_apikey(ctx):
    if not (
        isinstance(ctx.obj["AUTH"].auth_client(), PlanetLegacyAuthClient)
        or isinstance(ctx.obj["AUTH"].auth_client(), StaticApiKeyAuthClient)
    ):
        raise click.ClickException(
            "This command can only be used with "
            f'"{PlanetLegacyAuthClientConfig.meta()["client_type"]}" or "{StaticApiKeyAuthClientConfig.meta()["client_type"]}" '
            "type auth profiles. "
            f'The current profile "{ctx.obj["AUTH"].profile_name()}" is of type "{ctx.obj["AUTH"].auth_client()._auth_client_config.meta()["client_type"]}".'
        )


@click.group("legacy", invoke_without_command=True)
@click.pass_context
def cmd_pllegacy(ctx):
    """
    Planet legacy authentication commands.
    """
    if ctx.invoked_subcommand is None:
        click.echo(ctx.get_help())
        sys.exit(0)

    _check_client_type_pllegacy(ctx)


@cmd_pllegacy.command("login")
@opt_password(hidden=False)
@opt_username(hidden=False)
@opt_sops
@opt_yes_no
@click.pass_context
def cmd_pllegacy_login(ctx, username, password, sops, yes):
    """
    Perform an initial login using Planet's legacy authentication interfaces.
    """
    _check_client_type_pllegacy(ctx)
    current_auth_context = ctx.obj["AUTH"]
    current_auth_context.login(
        allow_tty_prompt=True,
        username=username,
        password=password,
    )
    print("Login succeeded.")  # Errors should throw.
    post_login_cmd_helper(override_auth_context=current_auth_context, use_sops=sops, prompt_pre_selection=yes)


@cmd_pllegacy.command("print-api-key")
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_pllegacy_print_api_key(ctx):
    """
    Show the API Key used by the currently selected authentication profile.
    """
    # We also support StaticApiKeyAuthClient in some cases where the user
    # directly provides an API key.  Such clients can use the legacy API
    # key, but lack the ability to obtain one.  This helps in our transition
    # away from the legacy auth protocol.
    _check_client_type_pllegacy_or_apikey(ctx)

    # Since API keys are static, we support them in the client config
    # and not just in the token file.
    if isinstance(ctx.obj["AUTH"].auth_client(), PlanetLegacyAuthClient):
        api_key = ctx.obj["AUTH"].auth_client().config().legacy_api_key()
        if api_key:
            print(api_key)
            return
    if isinstance(ctx.obj["AUTH"].auth_client(), StaticApiKeyAuthClient):
        api_key = ctx.obj["AUTH"].auth_client().config().api_key()
        if api_key:
            print(api_key)
            return

    saved_token = FileBackedPlanetLegacyApiKey(api_key_file=ctx.obj["AUTH"].token_file_path())
    print(saved_token.legacy_api_key())


@cmd_pllegacy.command("print-access-token")
@click.pass_context
@recast_exceptions_to_click(AuthException, FileNotFoundError)
def cmd_pllegacy_print_access_token(ctx):
    """
    Show the legacy JWT currently held by the selected authentication profile.
    """
    _check_client_type_pllegacy(ctx)
    saved_token = FileBackedPlanetLegacyApiKey(api_key_file=ctx.obj["AUTH"].token_file_path())
    print(saved_token.legacy_jwt())



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import shutil
import unittest

import planet_auth
import planet_auth.storage_utils
from planet_auth.constants import AUTH_CONFIG_FILE_PLAIN

from planet_auth_utils.builtins import Builtins
from planet_auth_utils.constants import EnvironmentVariables
from planet_auth_utils.plauth_factory import PlanetAuthFactory, MissingArgumentException
from planet_auth_utils.plauth_user_config import PlanetAuthUserConfig
from planet_auth_utils.profile import Profile

from .builtins_test_impl import MockProductionEnv, MockStagingEnv

from tests.test_planet_auth_utils.util import tdata_resource_file_path, TestWithHomeDirProfiles

PROFILE1_NAME = "test_profile1"
PROFILE2_NAME = "test_profile2"
PROFILE3_NAME = "test_profile3"


class TestAuthClientContextInitHelpers(TestWithHomeDirProfiles, unittest.TestCase):
    def setUp(self):
        os.environ[EnvironmentVariables.AUTH_BUILTIN_PROVIDER] = (
            "tests.test_planet_auth_utils.unit.auth_utils.builtins_test_impl.BuiltinConfigurationProviderMockTestImpl"
        )
        Builtins._builtin = None  # Reset built-in state.
        self.setUp_testHomeDir()

        self.profile1_dir_path = self.mkProfileDir(PROFILE1_NAME)
        shutil.copy(
            tdata_resource_file_path("auth_client_configs/utest/static_api_key.json"),
            self.profile1_dir_path.joinpath(planet_auth.constants.AUTH_CONFIG_FILE_PLAIN),
        )

        self.profile2_dir_path = self.mkProfileDir(PROFILE2_NAME)
        shutil.copy(
            tdata_resource_file_path("auth_client_configs/utest/static_api_key.json"),
            self.profile2_dir_path.joinpath(planet_auth.constants.AUTH_CONFIG_FILE_PLAIN),
        )

        self.profile3_dir_path = self.mkProfileDir(PROFILE3_NAME)
        shutil.copy(
            tdata_resource_file_path("auth_client_configs/utest/static_api_key.json"),
            self.profile3_dir_path.joinpath(planet_auth.constants.AUTH_CONFIG_FILE_PLAIN),
        )

        self.under_test_storage_provider = planet_auth.storage_utils.ObjectStorageProvider._default_storage_provider()

    def tearDown(self) -> None:
        self.tearDown_testHomeDir()
        os.environ.pop(EnvironmentVariables.AUTH_PROFILE, None)
        os.environ.pop(EnvironmentVariables.AUTH_API_KEY, None)
        os.environ.pop(EnvironmentVariables.AUTH_CLIENT_ID, None)
        os.environ.pop(EnvironmentVariables.AUTH_CLIENT_SECRET, None)

    def write_test_conf_file(self, data):
        conf_file = PlanetAuthUserConfig(data=data)
        conf_file.save()

    def test_default_initialization(self):
        under_test = PlanetAuthFactory.initialize_auth_client_context()
        self.assertEqual(
            under_test.profile_name(), Builtins.dealias_builtin_profile(Builtins.builtin_default_profile_name())
        )
        # self.assertIsInstance(under_test.auth_client(), planet_auth.XXX)
        self.assertIsNotNone(under_test.token_file_path())

    def test_init_from_dot_planet_json_profile(self):
        self.write_test_conf_file(
            {
                EnvironmentVariables.AUTH_PROFILE: PROFILE3_NAME,
                EnvironmentVariables.AUTH_CLIENT_ID: "unit_test_conffile_client_id",  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_SECRET: "unit_test_conffile_client_secret",  # Lower priority
                EnvironmentVariables.AUTH_API_KEY: "PLAK_conffile_API_key",  # Lower priority
            }
        )

        under_test = PlanetAuthFactory.initialize_auth_client_context()
        self.assertEqual(under_test.profile_name(), PROFILE3_NAME)
        # self.assertIsInstance(under_test.auth_client(), planet_auth.XXX)
        self.assertIsNotNone(under_test.token_file_path())

    def test_init_from_dot_planet_json_m2m(self):
        self.write_test_conf_file(
            {
                # EnvironmentVariables.AUTH_PROFILE: PROFILE3_NAME, # Higher priority
                EnvironmentVariables.AUTH_CLIENT_ID: "unit_test_conffile_client_id",
                EnvironmentVariables.AUTH_CLIENT_SECRET: "unit_test_conffile_client_secret",
                EnvironmentVariables.AUTH_API_KEY: "PLAK_conffile_API_key",  # Lower priority
            }
        )

        under_test = PlanetAuthFactory.initialize_auth_client_context()
        self.assertEqual(under_test.profile_name(), "utest-m2m-unit_test_conffile_client_id")
        self.assertEqual(
            under_test._auth_client._auth_client_config.lazy_get("client_id"), "unit_test_conffile_client_id"
        )
        self.assertEqual(
            under_test._auth_client._auth_client_config.lazy_get("client_secret"), "unit_test_conffile_client_secret"
        )
        self.assertIsInstance(under_test.auth_client(), planet_auth.ClientCredentialsClientSecretAuthClient)
        self.assertIsNotNone(under_test.token_file_path())

    def test_init_from_dot_planet_json_api_key(self):
        self.write_test_conf_file(
            {
                # EnvironmentVariables.AUTH_PROFILE: PROFILE3_NAME, # Higher priority
                # EnvironmentVariables.AUTH_CLIENT_ID: "unit_test_conffile_client_id", # Higher priority
                # EnvironmentVariables.AUTH_CLIENT_SECRET: "unit_test_conffile_client_secret", # Higher priority
                EnvironmentVariables.AUTH_API_KEY: "PLAK_conffile_API_key",
                "key": "PLAK_conffile_API_key_backwards_compat",
            }
        )

        under_test = PlanetAuthFactory.initialize_auth_client_context()
        self.assertEqual(under_test.profile_name(), "_PL_API_KEY")
        self.assertIsInstance(under_test.auth_client(), planet_auth.StaticApiKeyAuthClient)
        self.assertEqual(under_test._auth_client._auth_client_config.lazy_get("api_key"), "PLAK_conffile_API_key")
        self.assertIsNone(under_test.token_file_path())

    def test_init_from_dot_planet_json_api_key_backwards_compatibility(self):
        self.write_test_conf_file(
            {
                # We also accept the config key used by older versions of the SDK
                "key": "PLAK_conffile_API_key_backwards_compat",
            }
        )

        under_test = PlanetAuthFactory.initialize_auth_client_context()
        self.assertEqual(under_test.profile_name(), "_PL_API_KEY")
        self.assertIsInstance(under_test.auth_client(), planet_auth.StaticApiKeyAuthClient)
        self.assertEqual(
            under_test._auth_client._auth_client_config.lazy_get("api_key"), "PLAK_conffile_API_key_backwards_compat"
        )
        self.assertIsNone(under_test.token_file_path())

    def test_init_from_env_profile(self):
        self.write_test_conf_file(
            {
                EnvironmentVariables.AUTH_PROFILE: PROFILE3_NAME,  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_ID: "unit_test_conffile_client_id",  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_SECRET: "unit_test_conffile_client_secret",  # Lower priority
                EnvironmentVariables.AUTH_API_KEY: "PLAK_conffile_API_key",  # Lower priority
            }
        )
        os.environ[EnvironmentVariables.AUTH_PROFILE] = PROFILE1_NAME
        os.environ[EnvironmentVariables.AUTH_CLIENT_ID] = "unit_test_env_client_id"  # Lower priority
        os.environ[EnvironmentVariables.AUTH_CLIENT_SECRET] = "unit_test_env_client_secret"  # Lower priority
        os.environ[EnvironmentVariables.AUTH_API_KEY] = "PLAK_env_API_key"  # Lower priority

        under_test = PlanetAuthFactory.initialize_auth_client_context()
        self.assertEqual(under_test.profile_name(), PROFILE1_NAME)
        # self.assertIsInstance(under_test.auth_client(), planet_auth.XXX)
        self.assertIsNotNone(under_test.token_file_path())

    def test_init_from_env_oauth_m2m(self):
        self.write_test_conf_file(
            {
                # EnvironmentVariables.AUTH_PROFILE: PROFILE3_NAME,  # Higher priority
                EnvironmentVariables.AUTH_CLIENT_ID: "unit_test_conffile_client_id",  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_SECRET: "unit_test_conffile_client_secret",  # Lower priority
                EnvironmentVariables.AUTH_API_KEY: "PLAK_conffile_API_key",  # Lower priority
            }
        )
        os.environ.pop(EnvironmentVariables.AUTH_PROFILE, None)  # Higher priority
        os.environ[EnvironmentVariables.AUTH_CLIENT_ID] = "unit_test_env_client_id"
        os.environ[EnvironmentVariables.AUTH_CLIENT_SECRET] = "unit_test_env_client_secret"
        os.environ[EnvironmentVariables.AUTH_API_KEY] = "PLAK_env_API_key"  # Lower priority

        under_test = PlanetAuthFactory.initialize_auth_client_context()
        self.assertEqual(under_test.profile_name(), "utest-m2m-unit_test_env_client_id")
        self.assertEqual(under_test._auth_client._auth_client_config.lazy_get("client_id"), "unit_test_env_client_id")
        self.assertEqual(
            under_test._auth_client._auth_client_config.lazy_get("client_secret"), "unit_test_env_client_secret"
        )
        self.assertIsInstance(under_test.auth_client(), planet_auth.ClientCredentialsClientSecretAuthClient)
        self.assertIsNotNone(under_test.token_file_path())

    def test_init_from_env_api_key(self):
        self.write_test_conf_file(
            {
                # EnvironmentVariables.AUTH_PROFILE: PROFILE3_NAME,  # Higher priority
                # EnvironmentVariables.AUTH_CLIENT_ID: "unit_test_conffile_client_id",  # Higher priority
                # EnvironmentVariables.AUTH_CLIENT_SECRET: "unit_test_conffile_client_secret",  # Higher priority
                EnvironmentVariables.AUTH_API_KEY: "PLAK_conffile_API_key",  # Lower priority
            }
        )
        os.environ.pop(EnvironmentVariables.AUTH_PROFILE, None)  # Higher priority
        os.environ.pop(EnvironmentVariables.AUTH_CLIENT_ID, None)  # Higher priority
        os.environ.pop(EnvironmentVariables.AUTH_CLIENT_SECRET, None)  # Higher priority
        os.environ[EnvironmentVariables.AUTH_API_KEY] = "PLAK_env_API_key"

        under_test = PlanetAuthFactory.initialize_auth_client_context()
        self.assertEqual(under_test.profile_name(), "_PL_API_KEY")
        self.assertIsInstance(under_test.auth_client(), planet_auth.StaticApiKeyAuthClient)
        self.assertEqual(under_test._auth_client._auth_client_config.lazy_get("api_key"), "PLAK_env_API_key")
        self.assertIsNone(under_test.token_file_path())

    def test_init_from_explicit_profile(self):
        # Conflicting config from the env that should be ignored:
        self.write_test_conf_file(
            {
                EnvironmentVariables.AUTH_PROFILE: PROFILE3_NAME,  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_ID: "unit_test_conffile_client_id",  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_SECRET: "unit_test_conffile_client_secret",  # Lower priority
                EnvironmentVariables.AUTH_API_KEY: "PLAK_conffile_API_key",  # Lower priority
            }
        )
        os.environ[EnvironmentVariables.AUTH_PROFILE] = PROFILE1_NAME
        os.environ[EnvironmentVariables.AUTH_CLIENT_ID] = "unit_test_env_client_id"
        os.environ[EnvironmentVariables.AUTH_CLIENT_SECRET] = "unit_test_env_client_secret"
        os.environ[EnvironmentVariables.AUTH_API_KEY] = "PLAK_env_API_key"

        under_test = PlanetAuthFactory.initialize_auth_client_context(auth_profile_opt=PROFILE2_NAME)
        self.assertEqual(under_test.profile_name(), PROFILE2_NAME)
        # self.assertIsInstance(under_test.auth_client(), planet_auth.XXX)
        self.assertIsNotNone(under_test.token_file_path())

    def test_init_from_explicit_profile_unknown(self):
        # Conflicting config from the env that should be ignored:
        self.write_test_conf_file(
            {
                EnvironmentVariables.AUTH_PROFILE: PROFILE3_NAME,  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_ID: "unit_test_conffile_client_id",  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_SECRET: "unit_test_conffile_client_secret",  # Lower priority
                EnvironmentVariables.AUTH_API_KEY: "PLAK_conffile_API_key",  # Lower priority
            }
        )
        os.environ[EnvironmentVariables.AUTH_PROFILE] = PROFILE1_NAME
        os.environ[EnvironmentVariables.AUTH_CLIENT_ID] = "unit_test_env_client_id"
        os.environ[EnvironmentVariables.AUTH_CLIENT_SECRET] = "unit_test_env_client_secret"
        os.environ[EnvironmentVariables.AUTH_API_KEY] = "PLAK_env_API_key"

        with self.assertRaises(FileNotFoundError):  # as context:
            PlanetAuthFactory.initialize_auth_client_context(auth_profile_opt="Unit-Test-Unknown-Profile")

    def test_init_from_explicit_oauth_m2m(self):
        # Conflicting config from the env that should be ignored:
        self.write_test_conf_file(
            {
                EnvironmentVariables.AUTH_PROFILE: PROFILE3_NAME,  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_ID: "unit_test_conffile_client_id",  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_SECRET: "unit_test_conffile_client_secret",  # Lower priority
                EnvironmentVariables.AUTH_API_KEY: "PLAK_conffile_API_key",  # Lower priority
            }
        )
        os.environ[EnvironmentVariables.AUTH_CLIENT_ID] = "unit_test_env_client_id"
        os.environ[EnvironmentVariables.AUTH_CLIENT_SECRET] = "unit_test_env_client_secret"
        os.environ[EnvironmentVariables.AUTH_API_KEY] = "PLAK_env_API_key"

        mock_client_id = "test_client_id"
        mock_client_secret = "test_client_secret"
        under_test = PlanetAuthFactory.initialize_auth_client_context(
            auth_client_id_opt=mock_client_id,
            auth_client_secret_opt=mock_client_secret,
        )
        self.assertEqual(under_test.profile_name(), f"utest-m2m-{mock_client_id}")
        self.assertEqual(under_test._auth_client._auth_client_config.lazy_get("client_id"), "test_client_id")
        self.assertEqual(under_test._auth_client._auth_client_config.lazy_get("client_secret"), "test_client_secret")
        self.assertIsInstance(under_test.auth_client(), planet_auth.ClientCredentialsClientSecretAuthClient)
        self.assertIsNotNone(under_test.token_file_path())

    def test_init_from_explicit_api_key(self):
        # Conflicting config from the env that should be ignored:
        self.write_test_conf_file(
            {
                EnvironmentVariables.AUTH_PROFILE: PROFILE3_NAME,  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_ID: "unit_test_conffile_client_id",  # Lower priority
                EnvironmentVariables.AUTH_CLIENT_SECRET: "unit_test_conffile_client_secret",  # Lower priority
                EnvironmentVariables.AUTH_API_KEY: "PLAK_conffile_API_key",  # Lower priority
            }
        )
        os.environ[EnvironmentVariables.AUTH_PROFILE] = PROFILE1_NAME
        os.environ[EnvironmentVariables.AUTH_CLIENT_ID] = "unit_test_env_client_id"
        os.environ[EnvironmentVariables.AUTH_CLIENT_SECRET] = "unit_test_env_client_secret"
        os.environ[EnvironmentVariables.AUTH_API_KEY] = "PLAK_env_API_key"

        under_test = PlanetAuthFactory.initialize_auth_client_context(auth_api_key_opt="PLAK_TestApiKey")
        self.assertEqual(under_test.profile_name(), "_PL_API_KEY")
        self.assertEqual(under_test._auth_client._auth_client_config.lazy_get("api_key"), "PLAK_TestApiKey")
        self.assertIsInstance(under_test.auth_client(), planet_auth.StaticApiKeyAuthClient)
        self.assertIsNone(under_test.token_file_path())

    def test_token_file_override(self):
        default_under_test = PlanetAuthFactory.initialize_auth_client_context(
            auth_profile_opt=PROFILE1_NAME,
        )
        default_profile_token_file_path = default_under_test.token_file_path()
        self.assertIsNotNone(default_profile_token_file_path)

        requested_override_token_file_path = default_profile_token_file_path.with_suffix(".override-filename")
        override_under_test = PlanetAuthFactory.initialize_auth_client_context(
            auth_profile_opt=PROFILE1_NAME, token_file_opt=str(requested_override_token_file_path)
        )
        self.assertEqual(requested_override_token_file_path, override_under_test.token_file_path())

    def test_in_memory_token_file(self):
        token_saved_to_disk_under_test = PlanetAuthFactory.initialize_auth_client_context(
            auth_profile_opt=PROFILE1_NAME, save_token_file=True
        )
        self.assertIsNotNone(token_saved_to_disk_under_test.token_file_path())

        token_in_memory_only_under_test = PlanetAuthFactory.initialize_auth_client_context(
            auth_profile_opt=PROFILE1_NAME, save_token_file=False
        )
        self.assertIsNone(token_in_memory_only_under_test.token_file_path())

    def test_save_profile_saves_when_true(self):
        utest_profile_name = "new-utest-profile-save"
        client_conf_storage_path = Profile.get_profile_file_path(
            profile=utest_profile_name, filename=AUTH_CONFIG_FILE_PLAIN
        )
        self.assertFalse(self.under_test_storage_provider.obj_exists(client_conf_storage_path))

        _ = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config={
                "client_type": "oidc_client_credentials_secret",
                "auth_server": "https://login-utest.planet.com/",
                "audiences": ["https://utest.planet.com/"],
                "client_id": "__UTEST_CLIENT_ID__",
                "client_secret": "__UTEST_CLIENT_SECRET__",
            },
            save_token_file=True,
            save_profile_config=True,
            profile_name=utest_profile_name,
        )
        self.assertTrue(
            self.under_test_storage_provider.obj_exists(
                Profile.get_profile_file_path(profile=utest_profile_name, filename=AUTH_CONFIG_FILE_PLAIN)
            )
        )

    def test_save_profile_does_not_save_when_false(self):
        utest_profile_name = "new-utest-profile-nosave"
        client_conf_storage_path = Profile.get_profile_file_path(
            profile=utest_profile_name, filename=AUTH_CONFIG_FILE_PLAIN
        )
        self.assertFalse(self.under_test_storage_provider.obj_exists(client_conf_storage_path))

        _ = PlanetAuthFactory.initialize_auth_client_context_from_custom_config(
            client_config={
                "client_type": "oidc_client_credentials_secret",
                "auth_server": "https://login-utest.planet.com/",
                "audiences": ["https://utest.planet.com/"],
                "client_id": "__UTEST_CLIENT_ID__",
                "client_secret": "__UTEST_CLIENT_SECRET__",
            },
            save_token_file=False,
            save_profile_config=False,
            profile_name=utest_profile_name,
        )
        self.assertFalse(
            self.under_test_storage_provider.obj_exists(
                Profile.get_profile_file_path(profile=utest_profile_name, filename=AUTH_CONFIG_FILE_PLAIN)
            )
        )


ISSUER_PRIORITIES = ["_trusted"]
VALID_PRIMARY_CONFIGS = [
    {
        "auth_server": "https://login-fake.planet.com/oauth2/auth_server_id",
        "audiences": ["https://api.staging.planet-labs.com/"],
    }
]
VALID_SECONDARY_CONFIGS = [
    {
        "auth_server": "https://login-fake.planet.com/oauth2/deprecated_auth_server_id",
        "audiences": ["https://api.staging.planet-labs.com/"],
    }
]
INVALID_PRIMARY_CONFIGS = [{"auth_server": "https://login-fake.planet.com/oauth2/auth_server_id"}]


class TestResourceServerValidatorInitHelper(TestWithHomeDirProfiles, unittest.TestCase):
    def setUp(self):
        os.environ[EnvironmentVariables.AUTH_BUILTIN_PROVIDER] = (
            "tests.test_planet_auth_utils.unit.auth_utils.builtins_test_impl.BuiltinConfigurationProviderMockTestImpl"
        )
        Builtins._builtin = None  # Reset built-in state.
        self.setUp_testHomeDir()
        os.environ.pop(EnvironmentVariables.AUTH_PROFILE, None)
        os.environ.pop(EnvironmentVariables.AUTH_API_KEY, None)
        os.environ.pop(EnvironmentVariables.AUTH_CLIENT_ID, None)
        os.environ.pop(EnvironmentVariables.AUTH_CLIENT_SECRET, None)

    def tearDown(self) -> None:
        self.tearDown_testHomeDir()

    def _check_validator_equality(self, validator_1, validator_2):
        validator_1_vars = vars(validator_1)
        validator_2_vars = vars(validator_2)

        # Check the issuers
        for priority in ISSUER_PRIORITIES:
            self.assertEqual(validator_1_vars[priority].keys(), validator_2_vars[priority].keys())

    def test_staging(self):
        validator = planet_auth.OidcMultiIssuerValidator.from_auth_server_configs(
            MockStagingEnv.TRUSTED_OAUTH_AUTHORITIES,
        )
        test_validator = PlanetAuthFactory.initialize_resource_server_validator("STAGING")
        self._check_validator_equality(validator, test_validator)

    def test_prod(self):
        validator = planet_auth.OidcMultiIssuerValidator.from_auth_server_configs(
            MockProductionEnv.TRUESTED_OAUTH_AUTHORITIES,
        )
        test_validator = PlanetAuthFactory.initialize_resource_server_validator("PRODUCTION")
        self._check_validator_equality(validator, test_validator)

    def test_custom_valid_configs(self):
        test_validator = PlanetAuthFactory.initialize_resource_server_validator("CUSTOM", VALID_PRIMARY_CONFIGS)
        test_validator_vars = vars(test_validator)

        # Check if the issuers match the passed custom configuration
        for priority, config in zip(ISSUER_PRIORITIES, [VALID_PRIMARY_CONFIGS, VALID_SECONDARY_CONFIGS]):
            issuers = list(test_validator_vars[priority].keys())
            self.assertEqual(len(issuers), 1)
            self.assertEqual(issuers[0], config[0]["auth_server"])

    def test_custom_invalid_configs(self):
        with self.assertRaises(planet_auth.AuthException) as context:
            PlanetAuthFactory.initialize_resource_server_validator("CUSTOM", INVALID_PRIMARY_CONFIGS)
        self.assertEqual(
            str(context.exception),
            "Auth Providers used for OIDC token validation must have the audiences configuration value set.",
        )

    def test_custom_null_configs(self):
        with self.assertRaises(MissingArgumentException) as context:
            PlanetAuthFactory.initialize_resource_server_validator("CUSTOM", None)
        self.assertEqual(
            str(context.exception),
            "Custom or unknown environment was selected, but trusted_auth_server_configs was not supplied.",
        )

    def test_not_custom_with_configs(self):
        with self.assertWarns(UserWarning) as context:
            test_validator = PlanetAuthFactory.initialize_resource_server_validator("STAGING", VALID_PRIMARY_CONFIGS)
        self.assertEqual(
            len(context.warnings), 1, f"Expected exactly 1 warning, but got {len(context.warnings)} warnings."
        )
        self.assertEqual(
            str(context.warnings[0].message),
            "Custom environment not selected; trusted_auth_server_configs will be ignored in favor of the built in configuration for STAGING.",
        )

        validator = planet_auth.OidcMultiIssuerValidator.from_auth_server_configs(
            MockStagingEnv.TRUSTED_OAUTH_AUTHORITIES,
        )
        self._check_validator_equality(validator, test_validator)

    def test_environment_case_insensitivity(self):
        validator = planet_auth.OidcMultiIssuerValidator.from_auth_server_configs(
            MockStagingEnv.TRUSTED_OAUTH_AUTHORITIES,
        )
        test_validator = PlanetAuthFactory.initialize_resource_server_validator("sTaGinG")
        self._check_validator_equality(validator, test_validator)

    def test_unknown_environment(self):
        invalid_environment = "Omicron Persei 8"
        with self.assertRaises(ValueError) as context:
            PlanetAuthFactory.initialize_resource_server_validator(invalid_environment)
        self.assertEqual(
            str(context.exception),
            f"Passed environment must be one of {Builtins.builtin_environment_names()}. Instead, got: {invalid_environment.upper()}",
        )

    def test_unset_environment_1(self):
        with self.assertRaises(ValueError) as context:
            PlanetAuthFactory.initialize_resource_server_validator("")
        self.assertEqual(
            str(context.exception),
            f"Passed environment must be one of {Builtins.builtin_environment_names()}.",
        )

    def test_unset_environment_2(self):
        with self.assertRaises(ValueError) as context:
            PlanetAuthFactory.initialize_resource_server_validator(None)
        self.assertEqual(
            str(context.exception),
            f"Passed environment must be one of {Builtins.builtin_environment_names()}.",
        )



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import unittest

from planet_auth.none.noop_auth import NoOpCredential, NoOpAuthClientConfig, NoOpAuthClient

TEST_CREDENTIAL_FILE = "/some/path/that/does/not/exist.json"


class NoOpCredentialTest(unittest.TestCase):
    def test_noop_cred(self):
        under_test = NoOpCredential()
        under_test.set_path(TEST_CREDENTIAL_FILE)

        under_test.load()  # Should do nothing.

        under_test.save()
        self.assertFalse(pathlib.Path(TEST_CREDENTIAL_FILE).exists())


class NoOpAuthClientTest(unittest.TestCase):
    def setUp(self):
        self.under_test = NoOpAuthClient(NoOpAuthClientConfig())

    def test_login(self):
        cred = self.under_test.login()
        self.assertIsInstance(cred, NoOpCredential)

    def test_refresh(self):
        cred = self.under_test.refresh("refresh_token", requested_scopes=[])
        self.assertIsInstance(cred, NoOpCredential)

    def test_validate_access_token(self):
        results = self.under_test.validate_access_token_remote("test_token")
        self.assertEqual({}, results)

    def test_validate_access_token_local(self):
        results = self.under_test.validate_access_token_local("test_token", "test_audience")
        self.assertEqual({}, results)

    def test_validate_id_token(self):
        results = self.under_test.validate_id_token_remote("test_token")
        self.assertEqual({}, results)

    def test_validate_id_token_local(self):
        results = self.under_test.validate_id_token_local("test_token")
        self.assertEqual({}, results)

    def test_validate_refresh_token(self):
        results = self.under_test.validate_refresh_token_remote("test_token")
        self.assertEqual({}, results)

    def test_revoke_access_token(self):
        # test passes if it doesn't raise an exception
        self.under_test.revoke_access_token("test_token")

    def test_revoke_refresh_token(self):
        # test passes if it doesn't raise an exception
        self.under_test.revoke_refresh_token("test_token")

    def test_get_scopes(self):
        results = self.under_test.get_scopes()
        self.assertEqual([], results)



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import unittest

from requests.models import Response
from unittest import mock

from planet_auth.oidc.api_clients.jwks_api_client import JwksApiClient, JwksApiException

TEST_API_ENDPOINT = "https://blackhole.unittest.planet.com/api"
# Random snapshot of public keys from one of our Okta instances. We are
# not verifying keys in this test, so it doesn't matter that we can't generate
# tokens with these signers.
API_RESPONSE_VALID = {
    "keys": [
        {
            "kty": "RSA",
            "alg": "RS256",
            "kid": "WtxCYJa7oQqkLTwJYTfSeB1ZxUcY9h4tF7NL26Px1Tg",
            "use": "sig",
            "e": "AQAB",
            "n": "laSDEVkNhVp7XIvj6YYLELWstWefAR8it12K-qTDVsHCRDfOoPh97Yv0A3cVH6QPM_IoVTJ224yIQgpdTUykk3IN7Yem9WaZ0FC5AxjXWpBHDZpP5ZAjpBsR_r66pB2l_wDSDXZlN_k4VR4rbTjiGGsLqgY9J2QDLPyr5dormrO7E0inPuwLor2KTJkafSZd6AjifUkYS6zJQLvXenmD_7QTRVnYX2R07lzwctykVAj6ppM9KwQcqiSy9x-HqHm2wikQ1KzqBmXdCMLXB-FCFV4SbmC94744Rmhdb85bonX7TjZfvrurzBpO9ZhpSOeYXIDBwuJA4bsYYw-o6PYZaw",
        },
        {
            "kty": "RSA",
            "alg": "RS256",
            "kid": "S8fLvVhqvvxLxiwBdT7FXfIPGkV-I6LuRCwsaUsFqB4",
            "use": "sig",
            "e": "AQAB",
            "n": "l7Zho6s9npWRBX0590Vum9y4H5IIY2oTG-bJfw6OEIsMLDK8NW9jcFtFGsB7vuUvnPwiK9drI_ekApDZ8byfpIzBkJ9CIz3CU2xNZIaOEH8oU2Ywoviov6VBMDoSnoURGr57AYDnnuWsihgiZaH58ghmitbF_n9PPxhR62OP6M_fk8Dr7tDbtt4_mtVziLH8zDw0SrGFZZLhFmfmIGzoXxInnCpBCa6iVwqDjypB499-H2uIYyqiNoc-sR-CNUUFiRFZxZZeZDxkRm_M_gGVVbqCMgtzr8Y2z37hn2ZO1ylZc9TfOw2DY2dy4h7CKfjbhlAYn0Ckp6vxbopVhagY4Q",
        },
    ]
}
# API_RESPONSE_EMPTY_KEYSET = {"keys": []}
API_RESPONSE_INVALID_RESPONSE = {"bad": "payload"}


def mocked_response_ok(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(API_RESPONSE_VALID))
    return response


# def mocked_response_empty(request_url, **kwargs):
#    response = Response()
#    response.status_code = 200
#    response.headers['content-type'] = 'application/json'
#    response._content = str.encode(json.dumps(API_RESPONSE_EMPTY_KEYSET))
#    return response


def mocked_response_invalid(request_url, **kwargs):
    response = Response()
    response.status_code = 200
    response.headers["content-type"] = "application/json"
    response._content = str.encode(json.dumps(API_RESPONSE_INVALID_RESPONSE))
    return response


class JWKSApiClientTest(unittest.TestCase):
    @mock.patch("requests.sessions.Session.get", side_effect=mocked_response_ok)
    def test_fetch_keys_valid(self, mock_get):
        under_test = JwksApiClient(jwks_uri=TEST_API_ENDPOINT)
        keys = under_test.jwks_keys()
        self.assertEqual(API_RESPONSE_VALID["keys"], keys)

    @mock.patch("requests.sessions.Session.get", side_effect=mocked_response_invalid)
    def test_invalid(self, mock_get):
        under_test = JwksApiClient(jwks_uri=TEST_API_ENDPOINT)
        with self.assertRaises(JwksApiException):
            under_test.jwks_keys()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import cryptography.hazmat.primitives.serialization as crypto_serialization
import importlib.resources
import os
import socket
import sys

from contextlib import closing

from planet_auth.auth_client import AuthClientConfig


def is_interactive_shell() -> bool:
    return sys.stdin.isatty()


def is_not_interactive_shell() -> bool:
    return not sys.stdin.isatty()


def is_cicd() -> bool:
    # CI - GitHub
    # CI_COMMIT_SHA - GitLab
    return bool(os.getenv("CI") or os.getenv("CI_COMMIT_SHA"))


def tdata_resource_file_path(resource_file: str):
    file_path = importlib.resources.files("tests.test_planet_auth").joinpath("data/" + resource_file)
    return file_path


def load_auth_client_config(named_config):
    sops_path = tdata_resource_file_path("auth_client_configs/{}.sops.json".format(named_config))
    clear_path = tdata_resource_file_path("auth_client_configs/{}.json".format(named_config))
    if sops_path.is_file():
        conf_path = sops_path
    else:
        conf_path = clear_path
    return AuthClientConfig.from_file(conf_path)


def find_free_port():
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return s.getsockname()[1]


def load_rsa_private_key(key_file_path, password=None):
    with open(key_file_path, "rb") as key_file:
        if password:
            encoded_password = password.encode()
        else:
            encoded_password = None

        priv_key = crypto_serialization.load_pem_private_key(key_file.read(), password=encoded_password)
        if not priv_key:
            raise RuntimeError("Could not load private key from {}".format(key_file_path))

    return priv_key


def background(f):
    from functools import wraps

    @wraps(f)
    def wrapped(*args, **kwargs):
        loop = asyncio.get_event_loop()
        if callable(f):
            return loop.run_in_executor(None, f, *args, **kwargs)
        else:
            raise TypeError("Task must be a callable")

    return wrapped


def mock_sleep_skip(seconds):
    pass


class FreezeGunMockSleep:
    def __init__(self, frozen_time):
        self.frozen_time = frozen_time

    def mock_sleep(self, seconds):
        self.frozen_time.tick(seconds)



from planet_auth import Auth
import planet_auth_config  # type: ignore


def main():
    _legacy_client_config = {
        **planet_auth_config.Staging.LEGACY_AUTH_AUTHORITY,
        "client_type": "planet_legacy",
    }
    # Be careful where you save the token file
    auth_ctx = Auth.initialize_from_config_dict(client_config=_legacy_client_config, token_file="/saved_token.json")
    auth_ctx.login(allow_tty_prompt=True)


if __name__ == "__main__":
    main()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, Optional

from planet_auth.oidc.api_clients.api_client import OidcApiClient, OidcApiClientException


class DiscoveryApiException(OidcApiClientException):
    def __init__(self, message=None, raw_response=None):
        super().__init__(message, raw_response)


class DiscoveryApiClient(OidcApiClient):
    """
    Low level client for OIDC discovery.
    See https://openid.net/specs/openid-connect-discovery-1_0.html for details.
    """

    # TODO: Revisit if this is where I should cache. I did work
    #       on the JWKS client after this, and I think it is more mature.
    def __init__(self, discovery_uri: Optional[str] = None, auth_server: Optional[str] = None):
        """
        Create a new OIDC discovery API client.
        """
        if discovery_uri:
            d_uri = discovery_uri
        else:
            if auth_server:
                if auth_server.endswith("/"):
                    d_uri = auth_server + ".well-known/openid-configuration"
                else:
                    d_uri = auth_server + "/.well-known/openid-configuration"
            else:
                raise DiscoveryApiException("One of discovery_uri or auth_server must be provided")

        super().__init__(d_uri)
        self._oidc_discovery: Dict = None  # type: ignore

    def do_discovery(self) -> None:
        """
        Contact the discovery endpoint, download discovery information, and cache the
        results inside the client.
        """
        self._oidc_discovery = self._checked_get_json_response(None, None)

    def do_discovery_jit(self) -> None:
        """
        Contact the discovery endpoint and download and cache the results
        only if discovery had not previously been performed and cached.
        """
        if not self._oidc_discovery:
            self.do_discovery()

    def discovery(self) -> Dict:
        """
        Return the discovery information.  If the information was previously fetched,
        the cached information will be returned, and no network connection will be made.
        """
        self.do_discovery_jit()
        return self._oidc_discovery



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Dict, List, Optional


_NOOP_AUTH_CLIENT_CONFIG = {
    "client_type": "none",
}


class BuiltinConfigurationProviderInterface(ABC):
    """
    Interface to define what profiles are built-in.

    What auth configuration profiles are built-in is
    completely pluggable for users of the planet_auth and
    planet_auth_utils packages.  This is to support reuse
    in different deployments, or even support reuse by a
    different software stack all together.

    To inject built-in that override the coded in defaults,
    set the environment variable PL_AUTH_BUILTIN_CONFIG_PROVIDER
    to the module.classname of a class that implements this interface.

    Built-in profile names are expected to be all lowercase.

    Built-in trust environments are expected to be all uppercase.
    """

    @abstractmethod
    def builtin_client_authclient_config_dicts(self) -> Dict[str, dict]:
        """
        Return a dictionary of built-in AuthClient configuration
        dictionaries, keyed by a unique profile name.
        The returned dictionary values should be suitable for
        creating a functional configuration using
        `planet_auth.AuthClientConfig.config_from_dict`
        """

    @abstractmethod
    def builtin_client_profile_aliases(self) -> Dict[str, str]:
        """
        Return a dictionary profile aliases.  Aliases allow
        for a single built-in configuration to be referred to
        by multiple names.
        """

    @abstractmethod
    def builtin_default_profile_by_client_type(self) -> Dict[str, str]:
        """
        Return a dictionary of client types to default profile names for each client type.
        """

    @abstractmethod
    def builtin_default_profile(self) -> str:
        """
        Return the built-in default fallback auth profile name of last resort.
        """

    @abstractmethod
    def builtin_trust_environment_names(self) -> List[str]:
        """
        Return a list of the names of built-in trust environments.
        """

    @abstractmethod
    def builtin_trust_environments(self) -> Dict[str, Optional[List[dict]]]:
        """
        Return a dictionary of the trust environment configurations.
        Each item in the lists should be valid AuthClient config dictionary.
        """


class EmptyBuiltinProfileConstants(BuiltinConfigurationProviderInterface):
    BUILTIN_PROFILE_NAME_NONE = "none"

    NONE_AUTH_CLIENT_CONFIG = {
        "client_type": "none",
    }

    _builtin_profile_auth_client_configs = {
        BUILTIN_PROFILE_NAME_NONE: NONE_AUTH_CLIENT_CONFIG,
    }

    _builtin_profile_aliases: Dict[str, str] = {}

    _builtin_profile_default_by_client_type = {
        "none": BUILTIN_PROFILE_NAME_NONE,
    }
    _builtin_trust_realms: Dict[str, Optional[List[dict]]] = {
        "CUSTOM": None,
    }

    def builtin_client_authclient_config_dicts(self) -> Dict[str, dict]:
        return EmptyBuiltinProfileConstants._builtin_profile_auth_client_configs

    def builtin_client_profile_aliases(self) -> Dict[str, str]:
        return EmptyBuiltinProfileConstants._builtin_profile_aliases

    def builtin_default_profile_by_client_type(self) -> Dict[str, str]:
        return EmptyBuiltinProfileConstants._builtin_profile_default_by_client_type

    def builtin_default_profile(self) -> str:
        return EmptyBuiltinProfileConstants.BUILTIN_PROFILE_NAME_NONE

    def builtin_trust_environment_names(self) -> List[str]:
        return list(EmptyBuiltinProfileConstants._builtin_trust_realms.keys())

    def builtin_trust_environments(self) -> Dict[str, Optional[List[dict]]]:
        return EmptyBuiltinProfileConstants._builtin_trust_realms



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
from typing import List, Optional, Union

from planet_auth import CredentialRequestAuthenticator
from planet_auth.auth_client import AuthClientException
from planet_auth.credential import Credential
from planet_auth.oidc.auth_client import OidcAuthClientConfig, OidcAuthClient, FileBackedOidcCredential
from planet_auth.oidc.api_clients.api_client import EnricherPayloadType, EnricherReturnType
from planet_auth.request_authenticator import ForbiddenRequestAuthenticator


class OidcClientValidatorAuthClientConfig(OidcAuthClientConfig):
    """
    Configuration for a [planet_auth.OidcClientValidatorAuthClient][] AuthClient
    """

    INTERNAL_CLIENT_ID = "__OidcClientValidatorAuthClientConfig__"

    def __init__(self, **kwargs):
        if not "client_id" in kwargs:
            # Note: This is partly wrong.  This can work for local only
            #       validation, but for remote validation you usually need
            #       to auth and have a client ID. Rename this "local validator"?
            # Client ID is required in the base class, and pretty much all Oidc clients,
            # but we don't really have a need for it, since this AuthClient
            # class exists largely to paint a AuthClient veneer on the downloading
            # of JWKS pub keys and doing JWT validation.
            kwargs["client_id"] = self.INTERNAL_CLIENT_ID
        super().__init__(**kwargs)

    @classmethod
    def meta(cls):
        return {
            "client_type": "oidc_client_validator",
            "auth_client_class": OidcClientValidatorAuthClient,
            "display_name": "JWT Validator",
            "description": "Auth client that is only capable performing local validation of access tokens."
            " Cannot be used to make authenticated network calls.",
            "config_hints": super().meta().get("config_hints"),  # The superclasses have all we need.
        }


class OidcClientValidatorAuthClient(OidcAuthClient):
    """
    The OidcClientValidatorAuthClient is an OAuth2/OIDC client that does
    not have an identity of its own. It cannot perform a login or obtain
    access or ID tokens, and it cannot call out to resource servers
    covered by the issuer's audiences on behalf of users or itself. Instead,
    this OAuth auth client type exists for programs that need to validate
    tokens than have been presented to them. This is useful for resource
    servers themselves when authenticating or authorizing clients.

    The implementation is essentially the [planet_auth.OidcAuthClient][]
    base class with outbound client capabilities disabled.

    It should be noted that the other [planet_auth.OidcAuthClient][]
    derived classes that implement various OAuth flows can also be
    used to validate clients when inbound and outbound connections
    fall under the same token issuer realm of trust.  But, for use
    cases where only the validation of incoming requests is required,
    this class is suitable and does not require the allocation of a
    client ID.
    """

    def __init__(self, client_config: OidcClientValidatorAuthClientConfig):
        super().__init__(client_config)
        # Not used at this time
        # self._client_validator_client_config = client_config

    def _client_auth_enricher(self, raw_payload: EnricherPayloadType, audience: str) -> EnricherReturnType:
        return raw_payload, None

    def _oidc_flow_login(
        self,
        allow_open_browser: Optional[bool] = False,
        allow_tty_prompt: Optional[bool] = False,
        requested_scopes: Optional[List[str]] = None,
        requested_audiences: Optional[List[str]] = None,
        extra: Optional[dict] = None,
        **kwargs,
    ) -> FileBackedOidcCredential:
        raise AuthClientException(message="OIDC Client Validator auth client cannot perform a login.")

    def refresh(
        self, refresh_token: str, requested_scopes: List[str] = None, extra: Optional[dict] = None
    ) -> FileBackedOidcCredential:
        raise AuthClientException(message="OIDC Client Validator auth client cannot refresh credentials.")

    def default_request_authenticator(
        self, credential: Union[pathlib.Path, Credential]
    ) -> CredentialRequestAuthenticator:
        # return SimpleInMemoryRequestAuthenticator(token_body=None)  # Kinder and gentler failure.
        return ForbiddenRequestAuthenticator()

    def can_login_unattended(self) -> bool:
        return False
        # raise AuthClientException(message="OIDC Client Validator auth client cannot perform a login.")



import planet_auth_utils


def main():
    auth_ctx = planet_auth_utils.PlanetAuthFactory.initialize_auth_client_context(auth_profile_opt="my-custom-profile")
    print("Auth context initialized from on disk profile {}".format(auth_ctx.profile_name()))


if __name__ == "__main__":
    main()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pathlib
import unittest
from typing import Union
from unittest import mock
from unittest.mock import MagicMock

import pytest as pytest

from planet_auth.auth_client import AuthClientConfigException, AuthClientException
from planet_auth.credential import Credential
from planet_auth.oidc.api_clients.api_client import EnricherPayloadType, EnricherReturnType
from planet_auth.oidc.auth_client import (
    OidcAuthClient,
    OidcAuthClientConfig,
)
from planet_auth.oidc.auth_client_with_client_pubkey import OidcAuthClientWithPubKeyClientConfig
from planet_auth.oidc.auth_client_with_client_secret import OidcAuthClientWithClientSecretClientConfig
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.request_authenticator import CredentialRequestAuthenticator, SimpleInMemoryRequestAuthenticator
from tests.test_planet_auth.util import tdata_resource_file_path

TEST_CLIENT_ID = "_FAKE_CLIENT_ID_"
TEST_CLIENT_SECRET = "_FAKE_CLIENT_SECRET_"
TEST_ACCESS_TOKEN = "_FAKE_ACCESS_TOKEN_"
TEST_TOKEN_FILE = "/test/token.json"
TEST_FAKE_TOKEN_FILE_DATA = {
    "token_type": "Bearer",
    "expires_in": 3600,
    "access_token": "_dummy_access_token_",
    "scope": "offline_access profile openid",
    "refresh_token": "_dummy_refresh_token_",
    "id_token": "_dummy_id_token_",
}

TEST_AUTH_SERVER = "https://blackhole.unittest.planet.com/fake_authserver"
TEST_DISCOVERED_AUTH_SERVER_BASE = "https://auth.unittest.planet.com"
TEST_OVERRIDE_AUTH_SERVER_BASE = "https://auth_override.unittest.planet.com"
TEST_FAKE_OIDC_DISCOVERY = {
    "issuer": TEST_DISCOVERED_AUTH_SERVER_BASE,
    "authorization_endpoint": TEST_DISCOVERED_AUTH_SERVER_BASE + "/authorize",
    "token_endpoint": TEST_DISCOVERED_AUTH_SERVER_BASE + "/token",
    "userinfo_endpoint": TEST_DISCOVERED_AUTH_SERVER_BASE + "/userinfo",
    "registration_endpoint": "https://account.planet.com/oauth2/clients",
    "jwks_uri": TEST_DISCOVERED_AUTH_SERVER_BASE + "/jwks",
    "response_types_supported": [
        "code",
        "id_token",
        "code id_token",
        "code token",
        "id_token token",
        "code id_token token",
    ],
    "response_modes_supported": ["query", "fragment", "form_post", "okta_post_message"],
    "grant_types_supported": [
        "authorization_code",
        "implicit",
        "refresh_token",
        "password",
        "urn:ietf:params:oauth:grant-type:device_code",
    ],
    "subject_types_supported": ["public"],
    "id_token_signing_alg_values_supported": ["RS256"],
    "scopes_supported": ["openid", "profile", "email", "address", "phone", "offline_access", "device_sso"],
    "token_endpoint_auth_methods_supported": [
        "client_secret_basic",
        "client_secret_post",
        "client_secret_jwt",
        "private_key_jwt",
        "none",
    ],
    "claims_supported": [
        "iss",
        "ver",
        "sub",
        "aud",
        "iat",
        "exp",
        "jti",
        "auth_time",
        "amr",
        "idp",
        "nonce",
        "name",
        "nickname",
        "preferred_username",
        "given_name",
        "middle_name",
        "family_name",
        "email",
        "email_verified",
        "profile",
        "zoneinfo",
        "locale",
        "address",
        "phone_number",
        "picture",
        "website",
        "gender",
        "birthdate",
        "updated_at",
        "at_hash",
        "c_hash",
    ],
    "code_challenge_methods_supported": ["S256"],
    "introspection_endpoint": TEST_DISCOVERED_AUTH_SERVER_BASE + "/introspection",
    "introspection_endpoint_auth_methods_supported": [
        "client_secret_basic",
        "client_secret_post",
        "client_secret_jwt",
        "private_key_jwt",
        "none",
    ],
    "revocation_endpoint": TEST_DISCOVERED_AUTH_SERVER_BASE + "/revocation",
    "revocation_endpoint_auth_methods_supported": [
        "client_secret_basic",
        "client_secret_post",
        "client_secret_jwt",
        "private_key_jwt",
        "none",
    ],
    "end_session_endpoint": TEST_DISCOVERED_AUTH_SERVER_BASE + "/logout",
    "request_parameter_supported": True,
    "request_object_signing_alg_values_supported": [
        "HS256",
        "HS384",
        "HS512",
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
    ],
    "device_authorization_endpoint": TEST_DISCOVERED_AUTH_SERVER_BASE + "/device_auth",
}

TEST_FAKE_OIDC_DISCOVERY_NO_ENDPOINTS = {
    "issuer": TEST_DISCOVERED_AUTH_SERVER_BASE,
    "registration_endpoint": "https://account.planet.com/oauth2/clients",
    "response_types_supported": [
        "code",
        "id_token",
        "code id_token",
        "code token",
        "id_token token",
        "code id_token token",
    ],
    "response_modes_supported": ["query", "fragment", "form_post", "okta_post_message"],
    "grant_types_supported": [
        "authorization_code",
        "implicit",
        "refresh_token",
        "password",
        "urn:ietf:params:oauth:grant-type:device_code",
    ],
    "subject_types_supported": ["public"],
    "id_token_signing_alg_values_supported": ["RS256"],
    "scopes_supported": ["openid", "profile", "email", "address", "phone", "offline_access", "device_sso"],
    "token_endpoint_auth_methods_supported": [
        "client_secret_basic",
        "client_secret_post",
        "client_secret_jwt",
        "private_key_jwt",
        "none",
    ],
    "claims_supported": [
        "iss",
        "ver",
        "sub",
        "aud",
        "iat",
        "exp",
        "jti",
        "auth_time",
        "amr",
        "idp",
        "nonce",
        "name",
        "nickname",
        "preferred_username",
        "given_name",
        "middle_name",
        "family_name",
        "email",
        "email_verified",
        "profile",
        "zoneinfo",
        "locale",
        "address",
        "phone_number",
        "picture",
        "website",
        "gender",
        "birthdate",
        "updated_at",
        "at_hash",
        "c_hash",
    ],
    "code_challenge_methods_supported": ["S256"],
    "end_session_endpoint": TEST_DISCOVERED_AUTH_SERVER_BASE + "/logout",
    "request_parameter_supported": True,
    "request_object_signing_alg_values_supported": [
        "HS256",
        "HS384",
        "HS512",
        "RS256",
        "RS384",
        "RS512",
        "ES256",
        "ES384",
        "ES512",
    ],
    "device_authorization_endpoint": TEST_DISCOVERED_AUTH_SERVER_BASE + "/device/authorize",
}


def mocked_oidc_discovery(**kwargs):
    return TEST_FAKE_OIDC_DISCOVERY


def mocked_oidc_discovery_no_endpoints(**kwargs):
    return TEST_FAKE_OIDC_DISCOVERY_NO_ENDPOINTS


class OidcBaseTestHarnessClientConfig(OidcAuthClientConfig):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class OidcBaseTestHarnessAuthClient(OidcAuthClient):
    def __init__(self, client_config: OidcBaseTestHarnessClientConfig):
        super().__init__(client_config)
        self._test_client_config = client_config

    def _client_auth_enricher(self, raw_payload: EnricherPayloadType, audience: str) -> EnricherReturnType:
        return raw_payload, None

    def _oidc_flow_login(
        self, allow_open_browser, allow_tty_prompt, requested_scopes, requested_audiences, extra, **kwargs
    ) -> FileBackedOidcCredential:
        return FileBackedOidcCredential(data=TEST_FAKE_TOKEN_FILE_DATA)

    def default_request_authenticator(
        self, credential: Union[pathlib.Path, Credential]
    ) -> CredentialRequestAuthenticator:
        return SimpleInMemoryRequestAuthenticator(token_body=TEST_ACCESS_TOKEN)

    def can_login_unattended(self) -> bool:
        return False


def mocked_pem_load_error(key_data, **kwargs):
    return None


class AuthClientConfigBaseTest(unittest.TestCase):
    def test_authorization_callback_authorization_empty(self):
        under_test = OidcAuthClientConfig(auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID)
        under_test.check()
        self.assertIsNone(under_test.authorization_callback_acknowledgement_data())

    def test_authorization_callback_authorization_from_literal(self):
        under_test = OidcAuthClientConfig(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            authorization_callback_acknowledgement="test acknowledgement literal",
        )
        under_test.check()
        self.assertEqual(under_test.authorization_callback_acknowledgement_data(), "test acknowledgement literal")
        # Test coverage antidepressants: Coverate wants us to walk both branches of the lazy load
        self.assertEqual(under_test.authorization_callback_acknowledgement_data(), "test acknowledgement literal")

    def test_authorization_callback_authorization_from_file(self):
        resource_path = tdata_resource_file_path("resources/authorization_callback_acknowledgement.html")
        with open(resource_path, encoding="UTF-8") as resource_file:
            resource_file_str = resource_file.read()
            under_test = OidcAuthClientConfig(
                auth_server=TEST_AUTH_SERVER,
                client_id=TEST_CLIENT_ID,
                authorization_callback_acknowledgement_file=resource_path,
            )
            under_test.check()
            self.assertEqual(under_test.authorization_callback_acknowledgement_data(), resource_file_str)

    def test_authorization_callback_authorization_both_literal_and_file_set(self):
        # Code literal beats file
        resource_path = tdata_resource_file_path("resources/authorization_callback_acknowledgement.html")
        under_test = OidcAuthClientConfig(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            authorization_callback_acknowledgement="test acknowledgement literal",
            authorization_callback_acknowledgement_file=resource_path,
        )
        under_test.check()
        self.assertEqual(under_test.authorization_callback_acknowledgement_data(), "test acknowledgement literal")

    def test_required_config_fields(self):
        under_test = OidcAuthClientConfig(auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID)
        under_test.check()  # no throw

        under_test = OidcAuthClientConfig(client_id=TEST_CLIENT_ID)
        with self.assertRaises(AuthClientConfigException):
            under_test.check()

        under_test = OidcAuthClientConfig(auth_server=TEST_AUTH_SERVER)
        with self.assertRaises(AuthClientConfigException):
            under_test.check()

    def test_audiences_valid_value(self):
        under_test = OidcAuthClientConfig(auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, audiences=None)
        under_test.check()  # no throw

        under_test = OidcAuthClientConfig(auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, audiences=[])
        under_test.check()  # no throw

        under_test = OidcAuthClientConfig(auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, audiences=["aud1"])
        under_test.check()  # no throw

        # under_test = OidcAuthClientConfig(
        #     auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, audiences=""  # Falsy value is not list type
        # )
        # with self.assertRaises(AuthClientConfigException):
        #     under_test.check()

        under_test = OidcAuthClientConfig(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, audiences="value_is_not_list_type"
        )
        with self.assertRaises(AuthClientConfigException):
            under_test.check()

        under_test = OidcAuthClientConfig(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, audiences=["aud1", "aud2"]
        )
        with self.assertRaises(AuthClientConfigException):
            under_test.check()


class OidcAuthClientWithClientSecretClientConfigTest(unittest.TestCase):
    class TestConfigClass(OidcAuthClientWithClientSecretClientConfig):
        pass  # just test the ABC base.

    def test_secret_configured(self):
        under_test = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, client_secret=TEST_CLIENT_SECRET
        )
        under_test.check()
        self.assertEqual(TEST_CLIENT_SECRET, under_test.client_secret())

    def test_no_secret_configured(self):
        under_test = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
        )
        with self.assertRaises(AuthClientConfigException):
            under_test.check()


class OidcAuthClientWithPubKeyClientConfigTest(unittest.TestCase):
    class TestConfigClass(OidcAuthClientWithPubKeyClientConfig):
        pass  # just test the ABC base.

    @classmethod
    def setUpClass(cls):
        cls.privkey_password = "password"
        cls.privkey_file_path = tdata_resource_file_path("keys/keypair1_priv.test_pem")
        with open(cls.privkey_file_path, encoding="UTF-8") as key_file:
            cls.privkey_literal_str = key_file.read()

        cls.privkey_file_path_nopassword = tdata_resource_file_path("keys/keypair1_priv_nopassword.test_pem")
        with open(cls.privkey_file_path_nopassword, encoding="UTF-8") as key_file:
            cls.privkey_literal_str_nopassword = key_file.read()

    def _assert_rsa_keys_equal(self, key1, key2):
        # We are not validating the crypto libs. We assume if both keys
        # loaded without throwing and look similar, our code is working as
        # expected.
        self.assertEqual(key1.key_size, key2.key_size)

    def test_key_loads_from_literal_or_file(self):
        under_test_literal = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, client_privkey=self.privkey_literal_str_nopassword
        )
        under_test_literal.check()

        under_test_filebacked = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey_file=self.privkey_file_path_nopassword,
        )
        under_test_filebacked.check()

        self._assert_rsa_keys_equal(
            under_test_filebacked.client_privkey_data(), under_test_literal.client_privkey_data()
        )

    def test_key_loads_with_or_without_password_literal(self):
        under_test_nopw = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, client_privkey=self.privkey_literal_str_nopassword
        )
        under_test_nopw.check()

        under_test_pw = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey=self.privkey_literal_str,
            client_privkey_password=self.privkey_password,
        )
        under_test_pw.check()

        key_nopw = under_test_nopw.client_privkey_data()
        key_pw = under_test_pw.client_privkey_data()
        self._assert_rsa_keys_equal(key_pw, key_nopw)

    def test_key_loads_with_or_without_password_filebacked(self):
        under_test_nopw = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey_file=self.privkey_file_path_nopassword,
        )
        under_test_nopw.check()

        under_test_pw = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey_file=self.privkey_file_path,
            client_privkey_password=self.privkey_password,
        )
        under_test_pw.check()

        key_nopw = under_test_nopw.client_privkey_data()
        key_pw = under_test_pw.client_privkey_data()
        self._assert_rsa_keys_equal(key_pw, key_nopw)

    def test_bad_password_throws(self):
        under_test = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey_password=None,
            client_privkey=self.privkey_literal_str,
        )
        under_test.check()
        with self.assertRaises(AuthClientConfigException):
            under_test.client_privkey_data()

        under_test = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey_password="bad password",
            client_privkey=self.privkey_literal_str,
        )
        under_test.check()
        with self.assertRaises(AuthClientConfigException):
            under_test.client_privkey_data()

        under_test = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey_password=None,
            client_privkey_file=self.privkey_file_path,
        )
        under_test.check()
        with self.assertRaises(AuthClientConfigException):
            under_test.client_privkey_data()

        under_test = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey_password="bad password",
            client_privkey_file=self.privkey_file_path,
        )
        under_test.check()
        with self.assertRaises(AuthClientConfigException):
            under_test.client_privkey_data()

    @mock.patch("cryptography.hazmat.primitives.serialization.load_pem_private_key", mocked_pem_load_error)
    def test_unexpected_keyload_restult(self):
        under_test = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER, client_id=TEST_CLIENT_ID, client_privkey=self.privkey_literal_str_nopassword
        )
        under_test.check()
        with self.assertRaises(AuthClientConfigException):
            under_test.client_privkey_data()

        under_test = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey_file=self.privkey_file_path_nopassword,
        )
        under_test.check()
        with self.assertRaises(AuthClientConfigException):
            under_test.client_privkey_data()

    def test_no_key_configured(self):
        under_test = self.TestConfigClass(
            auth_server=TEST_AUTH_SERVER,
            client_id=TEST_CLIENT_ID,
            client_privkey=None,
            client_privkey_file=None,
            client_privkey_password=None,
        )
        with self.assertRaises(AuthClientConfigException):
            under_test.client_privkey_data()

    def test_lazy_load_only_once(self):
        # pylint: disable=fixme
        # FIXME: wrapping the object in this way doesn't seem to actually
        #   catch call counts for internal method calls.  Stepping through
        #   debugger seems to indicate proper behavior.
        under_test = MagicMock(
            wraps=self.TestConfigClass(
                auth_server=TEST_AUTH_SERVER,
                client_id=TEST_CLIENT_ID,
                client_privkey_file=self.privkey_file_path_nopassword,
            )
        )
        under_test.check()
        self.assertEqual(0, under_test._load_private_key.call_count)
        under_test.client_privkey_data()
        # self.assertEqual(1, under_test._load_private_key.call_count)
        under_test.client_privkey_data()
        # self.assertEqual(1, under_test._load_private_key.call_count)


@mock.patch(
    "planet_auth.oidc.api_clients.discovery_api_client.DiscoveryApiClient.discovery",
    side_effect=mocked_oidc_discovery,
)
class AuthClientBaseTest(unittest.TestCase):
    def setUp(self):
        self.defaults_under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                audiences=["audience_from_client_config_1"],
                # audiences=["audience_from_client_config_1", "audience_from_client_config_2"],
            )
        )
        self.oidc_test_credential = FileBackedOidcCredential(data=TEST_FAKE_TOKEN_FILE_DATA)

    def test_oidc_discovery_override_authorization(self, mocked_discovery):
        # OIDC discovery is intended to be JIT in the base class.
        # When an override is provided, 1) that discovery should not happen
        # (some OIDC providers don't offer discovery, so it would fail),
        # and 2) the override should be used, even if discovery was
        # previously successful for another endpoint.
        #
        # This should be tested for each endpoint we support:
        #   authorization_endpoint  (This one is an oddball, since it doesn't
        #                            make HTTP requests the same way)
        #   introspection_endpoint
        #   jwks_endpoint
        #   revocation_endpoint
        #   userinfo_endpoint
        #   token_endpoint

        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                authorization_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/authorize",
            )
        )
        api_client = under_test.authorization_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/authorize", api_client._authorization_uri)
        self.assertEqual(0, mocked_discovery.call_count)

        # discovery after the initial overridden client fetch (above) tests
        # that it's not clobbered by discovery
        under_test._discovery()
        api_client = under_test.authorization_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/authorize", api_client._authorization_uri)

        # Discovery before the initial fetch test that it's not preempted by
        # a prior discovery.
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                authorization_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/authorize",
            )
        )
        under_test._discovery()
        api_client = under_test.authorization_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/authorize", api_client._authorization_uri)

        # And finally, test default behavior:
        default_api_client = self.defaults_under_test.authorization_client()
        self.assertEqual(TEST_DISCOVERED_AUTH_SERVER_BASE + "/authorize", default_api_client._authorization_uri)

    def test_oidc_discovery_override_device_authorization(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                device_authorization_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/device_auth",
            )
        )
        api_client = under_test.device_authorization_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/device_auth", api_client._endpoint_uri)
        self.assertEqual(0, mocked_discovery.call_count)

        # discovery after the initial overridden client fetch (above) tests
        # that it's not clobbered by discovery
        under_test._discovery()
        api_client = under_test.device_authorization_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/device_auth", api_client._endpoint_uri)

        # Discovery before the initial fetch test that it's not preempted by
        # a prior discovery.
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                device_authorization_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/device_auth",
            )
        )
        under_test._discovery()
        api_client = under_test.device_authorization_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/device_auth", api_client._endpoint_uri)

        # And finally, test default behavior:
        default_api_client = self.defaults_under_test.device_authorization_client()
        self.assertEqual(TEST_DISCOVERED_AUTH_SERVER_BASE + "/device_auth", default_api_client._endpoint_uri)

    def test_oidc_discovery_override_introspection(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                introspection_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/introspection",
            )
        )
        api_client = under_test.introspection_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/introspection", api_client._endpoint_uri)
        self.assertEqual(0, mocked_discovery.call_count)

        # discovery after the initial overridden client fetch (above) tests
        # that it's not clobbered by discovery
        under_test._discovery()
        api_client = under_test.introspection_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/introspection", api_client._endpoint_uri)

        # Discovery before the initial fetch test that it's not preempted by
        # a prior discovery.
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                introspection_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/introspection",
            )
        )
        under_test._discovery()
        api_client = under_test.introspection_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/introspection", api_client._endpoint_uri)

        # And finally, test default behavior:
        default_api_client = self.defaults_under_test.introspection_client()
        self.assertEqual(TEST_DISCOVERED_AUTH_SERVER_BASE + "/introspection", default_api_client._endpoint_uri)

    def test_oidc_discovery_override_jwks(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                jwks_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/jwks",
            )
        )
        api_client = under_test.jwks_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/jwks", api_client._endpoint_uri)
        self.assertEqual(0, mocked_discovery.call_count)

        # discovery after the initial overridden client fetch (above) tests
        # that it's not clobbered by discovery
        under_test._discovery()
        api_client = under_test.jwks_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/jwks", api_client._endpoint_uri)

        # Discovery before the initial fetch test that it's not preempted by
        # a prior discovery.
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                jwks_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/jwks",
            )
        )
        under_test._discovery()
        api_client = under_test.jwks_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/jwks", api_client._endpoint_uri)

        # And finally, test default behavior:
        default_api_client = self.defaults_under_test.jwks_client()
        self.assertEqual(TEST_DISCOVERED_AUTH_SERVER_BASE + "/jwks", default_api_client._endpoint_uri)

    def test_oidc_discovery_override_revocation(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                revocation_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/revocation",
            )
        )
        api_client = under_test.revocation_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/revocation", api_client._endpoint_uri)
        self.assertEqual(0, mocked_discovery.call_count)

        # discovery after the initial overridden client fetch (above) tests
        # that it's not clobbered by discovery
        under_test._discovery()
        api_client = under_test.revocation_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/revocation", api_client._endpoint_uri)

        # Discovery before the initial fetch test that it's not preempted by
        # a prior discovery.
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                revocation_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/revocation",
            )
        )
        under_test._discovery()
        api_client = under_test.revocation_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/revocation", api_client._endpoint_uri)

        # And finally, test default behavior:
        default_api_client = self.defaults_under_test.revocation_client()
        self.assertEqual(TEST_DISCOVERED_AUTH_SERVER_BASE + "/revocation", default_api_client._endpoint_uri)

    def test_oidc_discovery_override_userinfo(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                userinfo_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/userinfo",
            )
        )
        api_client = under_test.userinfo_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/userinfo", api_client._endpoint_uri)
        self.assertEqual(0, mocked_discovery.call_count)

        # discovery after the initial overridden client fetch (above) tests
        # that it's not clobbered by discovery
        under_test._discovery()
        api_client = under_test.userinfo_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/userinfo", api_client._endpoint_uri)

        # Discovery before the initial fetch test that it's not preempted by
        # a prior discovery.
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                userinfo_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/userinfo",
            )
        )
        under_test._discovery()
        api_client = under_test.userinfo_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/userinfo", api_client._endpoint_uri)

        # And finally, test default behavior:
        default_api_client = self.defaults_under_test.userinfo_client()
        self.assertEqual(TEST_DISCOVERED_AUTH_SERVER_BASE + "/userinfo", default_api_client._endpoint_uri)

    def test_oidc_discovery_override_token(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                token_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/token",
            )
        )
        api_client = under_test.token_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/token", api_client._endpoint_uri)
        self.assertEqual(0, mocked_discovery.call_count)

        # discovery after the initial overridden client fetch (above) tests
        # that it's not clobbered by discovery
        under_test._discovery()
        api_client = under_test.token_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/token", api_client._endpoint_uri)

        # Discovery before the initial fetch test that it's not preempted by
        # a prior discovery.
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                token_endpoint=TEST_OVERRIDE_AUTH_SERVER_BASE + "/token",
            )
        )
        under_test._discovery()
        api_client = under_test.token_client()
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/token", api_client._endpoint_uri)

        # And finally, test default behavior:
        default_api_client = self.defaults_under_test.token_client()
        self.assertEqual(TEST_DISCOVERED_AUTH_SERVER_BASE + "/token", default_api_client._endpoint_uri)

    def test_issuer_discovery_override(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                issuer=TEST_OVERRIDE_AUTH_SERVER_BASE + "/override_issuer",
            )
        )
        self.assertEqual(TEST_OVERRIDE_AUTH_SERVER_BASE + "/override_issuer", under_test._issuer())
        self.assertEqual(0, mocked_discovery.call_count)

        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE, client_id=TEST_CLIENT_ID)
        )
        self.assertEqual(TEST_DISCOVERED_AUTH_SERVER_BASE, under_test._issuer())
        self.assertEqual(1, mocked_discovery.call_count)

        # second call doesn't re-do discovery
        under_test._issuer()
        self.assertEqual(1, mocked_discovery.call_count)

    def test_token_validator_created_once(self, mocked_discovery):
        under_test = self.defaults_under_test
        token_validator = under_test._token_validator()
        self.assertIsNotNone(token_validator)
        token_validator2 = under_test._token_validator()
        self.assertEqual(token_validator, token_validator2)

    #
    # The base client implementation of the following is pretty much
    # one liner pass through.  These nothing-burger tests are mostly
    # to goose the coverage targets, and would really only catch
    # gross errors in the pass through implementation. These are
    # better tested in end-to-end tests.
    #
    @mock.patch("planet_auth.oidc.api_clients.token_api_client.TokenApiClient.get_token_from_refresh")
    def test_refresh(self, mock_api_client, mocked_discovery):
        under_test = self.defaults_under_test
        credential = under_test.refresh(self.oidc_test_credential.refresh_token())
        self.assertIsInstance(credential, FileBackedOidcCredential)
        self.assertEqual(1, mock_api_client.call_count)

    @mock.patch("planet_auth.oidc.api_clients.introspect_api_client.IntrospectionApiClient.validate_access_token")
    def test_validate_access_token(self, mock_api_client, mocked_discovery):
        under_test = self.defaults_under_test
        under_test.validate_access_token_remote(self.oidc_test_credential.access_token())
        self.assertEqual(1, mock_api_client.call_count)

    @mock.patch("planet_auth.oidc.token_validator.TokenValidator.validate_token")
    def test_validate_access_token_local(self, mock_validate_token, mocked_discovery):
        under_test = self.defaults_under_test
        under_test.validate_access_token_local(
            self.oidc_test_credential.access_token(), required_audience="audience_from_validate_call"
        )
        self.assertEqual(1, mock_validate_token.call_count)
        args, kwargs = mock_validate_token.call_args
        self.assertEqual("audience_from_validate_call", kwargs.get("audience"))

    @mock.patch("planet_auth.oidc.token_validator.TokenValidator.validate_token")
    def test_validate_access_token_local_default_audience_from_config(self, mock_validate_token, mocked_discovery):
        under_test_with_username_password = self.defaults_under_test
        under_test_with_username_password.validate_access_token_local(self.oidc_test_credential.access_token())
        self.assertEqual(1, mock_validate_token.call_count)
        args, kwargs = mock_validate_token.call_args
        self.assertEqual("audience_from_client_config_1", kwargs.get("audience"))

    def test_validate_access_token_no_audience_fails(self, mock_validate_token):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                # audiences=["audience_from_client_config_1"],
                # audiences=["audience_from_client_config_1", "audience_from_client_config_2"],
            )
        )
        with self.assertRaises(AuthClientException):
            under_test.validate_access_token_local(self.oidc_test_credential.access_token())

    @pytest.mark.xfail(
        reason="This test check for a runtime error during token validation when the required audience is ambiguous."
        " Currently, the code prevents this from being a legal config, so rather than the runtime AuthClientException exception,"
        " a config time AuthClientConfigException exception is thrown."
        " Should configuration ever loosen generally, this test should remain and no longer be marked xfail"
        " since a runtime error should be thrown."
    )
    def test_validate_access_token_ambiguous_audience_fails(self, mock_validate_token):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                # audiences=["audience_from_client_config_1"],
                audiences=["audience_from_client_config_1", "audience_from_client_config_2"],
            )
        )
        with self.assertRaises(AuthClientException):
            under_test.validate_access_token_local(self.oidc_test_credential.access_token())

    @mock.patch("planet_auth.oidc.api_clients.introspect_api_client.IntrospectionApiClient.validate_id_token")
    def test_validate_id_token(self, mock_api_client, mocked_discovery):
        under_test = self.defaults_under_test
        under_test.validate_id_token_remote(self.oidc_test_credential.id_token())
        self.assertEqual(1, mock_api_client.call_count)

    @mock.patch("planet_auth.oidc.token_validator.TokenValidator.validate_id_token")
    def test_validate_id_token_local(self, mock_api_client, mocked_discovery):
        under_test = self.defaults_under_test
        under_test.validate_id_token_local(self.oidc_test_credential.id_token())
        self.assertEqual(1, mock_api_client.call_count)

    @mock.patch("planet_auth.oidc.api_clients.introspect_api_client.IntrospectionApiClient.validate_refresh_token")
    def test_validate_refresh_token(self, mock_api_client, mocked_discovery):
        under_test = self.defaults_under_test
        under_test.validate_refresh_token_remote(self.oidc_test_credential.refresh_token())
        self.assertEqual(1, mock_api_client.call_count)

    @mock.patch("planet_auth.oidc.api_clients.revocation_api_client.RevocationApiClient.revoke_access_token")
    def test_revoke_access_token(self, mock_api_client, mocked_discovery):
        under_test = self.defaults_under_test
        under_test.revoke_access_token(self.oidc_test_credential.access_token())
        self.assertEqual(1, mock_api_client.call_count)

    @mock.patch("planet_auth.oidc.api_clients.revocation_api_client.RevocationApiClient.revoke_refresh_token")
    def test_revoke_refresh_token(self, mock_api_client, mocked_discovery):
        under_test = self.defaults_under_test
        under_test.revoke_refresh_token(self.oidc_test_credential.refresh_token())
        self.assertEqual(1, mock_api_client.call_count)

    @mock.patch("planet_auth.oidc.api_clients.userinfo_api_client.UserinfoApiClient.userinfo_from_access_token")
    def test_userinfo_from_access_token(self, mock_api_client, mocked_discovery):
        under_test = self.defaults_under_test
        under_test.userinfo_from_access_token(self.oidc_test_credential.access_token())
        self.assertEqual(1, mock_api_client.call_count)

    def test_get_scopes(self, mocked_discovery):
        under_test = self.defaults_under_test
        test_scopes = under_test.get_scopes()
        self.assertEqual(TEST_FAKE_OIDC_DISCOVERY["scopes_supported"], test_scopes)


@mock.patch(
    "planet_auth.oidc.api_clients.discovery_api_client.DiscoveryApiClient.discovery",
    side_effect=mocked_oidc_discovery_no_endpoints,
)
class AuthClientBaseTestDiscoveryDoesNotAdvertiseEndpoints(unittest.TestCase):
    def setUp(self):
        self.defaults_under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
                audiences=["audience_from_client_config_1"],
                # audiences=["audience_from_client_config_1", "audience_from_client_config_2"],
            )
        )
        self.oidc_test_credential = FileBackedOidcCredential(data=TEST_FAKE_TOKEN_FILE_DATA)

    def test_oidc_discovery_no_authorization_endpoint(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
            )
        )
        with self.assertRaises(AuthClientException):
            under_test.authorization_client()

    def test_oidc_discovery_no_introspection_endpoint(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
            )
        )
        with self.assertRaises(AuthClientException):
            under_test.introspection_client()

    def test_oidc_discovery_no_jwks_endpoint(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
            )
        )
        with self.assertRaises(AuthClientException):
            under_test.jwks_client()

    def test_oidc_discovery_no_revocation_endpoint(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
            )
        )
        with self.assertRaises(AuthClientException):
            under_test.revocation_client()

    def test_oidc_discovery_no_userinfo_endpoint(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
            )
        )
        with self.assertRaises(AuthClientException):
            under_test.userinfo_client()

    def test_oidc_discovery_no_token_endpoint(self, mocked_discovery):
        under_test = OidcBaseTestHarnessAuthClient(
            OidcBaseTestHarnessClientConfig(
                auth_server=TEST_DISCOVERED_AUTH_SERVER_BASE,
                client_id=TEST_CLIENT_ID,
            )
        )
        with self.assertRaises(AuthClientException):
            under_test.token_client()



# Copyright 2024 Planet Labs PBC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import json
from abc import ABC, abstractmethod

import jwt.utils
import pathlib
import secrets
from typing import List, Optional, Union

import jwt
import time
import uuid

from planet_auth.storage_utils import (
    FileBackedJsonObjectException,
    ObjectStorageProvider_KeyType,
    ObjectStorageProvider,
)
from planet_auth.credential import Credential
from planet_auth.request_authenticator import CredentialRequestAuthenticator, ForbiddenRequestAuthenticator
from planet_auth.oidc.api_clients.api_client import EnricherPayloadType, EnricherReturnType
from planet_auth.oidc.oidc_credential import FileBackedOidcCredential
from planet_auth.oidc.api_clients.jwks_api_client import JwksApiClient
from planet_auth.oidc.api_clients.token_api_client import TokenApiException
from planet_auth.oidc.auth_client import OidcAuthClientConfig, OidcAuthClient

from tests.test_planet_auth.util import load_rsa_private_key, tdata_resource_file_path

_SCOPE_CLAIM_RFC8693 = "scope"  # See RFC 8693, 9068
_SCOPE_CLAIM_OKTA = "scp"  # Okta uses a different claim


class UtilTokenBuilderBase(ABC):
    """
    Build tokens locally for testing purposes.
    """

    @abstractmethod
    def encode(self, body, extra_headers):
        pass

    def __init__(self, issuer, audience, signing_key_id=None):
        self.signing_key_id = signing_key_id
        self.issuer = issuer
        self.audience = audience

    def _construct_oidc_access_token(self, username: str, ttl: int, extra_claims, remove_claims):
        # Build a fake JWT access token.  For test purposes, it doesn't matter
        # that it cannot be used anywhere.  While some fields are generic and standard
        # the full details of an access token are specific to the authorization
        # server and the application.
        now = int(time.time())
        unsigned_jwt = {
            "iss": self.issuer,
            "sub": username,
            "aud": self.audience,
            "iat": now,
            "exp": now + ttl,
            "jti": str(uuid.uuid4()),
        }
        unsigned_jwt.update(extra_claims)
        if remove_claims:
            for remove_claim in remove_claims:
                if remove_claim in unsigned_jwt:
                    unsigned_jwt.pop(remove_claim)

        headers = {}
        if self.signing_key_id:
            headers["kid"] = self.signing_key_id

        signed_jwt = self.encode(unsigned_jwt, extra_headers=headers)
        return signed_jwt

    def construct_oidc_access_token_okta(
        self, username: str, requested_scopes: list, ttl: int, extra_claims=None, remove_claims=None
    ):
        scope_claims = {"scp": requested_scopes}  # Claim used by Okta for granted scopes.
        if extra_claims:
            extra_claims.update(scope_claims)
        else:
            extra_claims = scope_claims
        return self._construct_oidc_access_token(
            username=username, ttl=ttl, extra_claims=extra_claims, remove_claims=remove_claims
        )

    def construct_oidc_access_token_rfc8693(
        self, username: str, requested_scopes: list, ttl: int, extra_claims=None, remove_claims=None
    ):
        scope_claims = {"scope": " ".join(requested_scopes)}  # RFC 8693 claim used for granted scopes
        if extra_claims:
            extra_claims.update(scope_claims)
        else:
            extra_claims = scope_claims
        return self._construct_oidc_access_token(
            username=username, ttl=ttl, extra_claims=extra_claims, remove_claims=remove_claims
        )

    def construct_oidc_id_token(self, client_id: str, ttl: int, extra_claims=None, remove_claims=None):
        # Build a fake JWT ID token.  For test purposes, it doesn't matter
        # that it cannot be used anywhere.
        now = int(time.time())
        unsigned_jwt = {
            "iss": self.issuer,
            "sub": client_id,
            "aud": client_id,
            "iat": now,
            "exp": now + ttl,
            "jti": str(uuid.uuid4()),
        }
        if extra_claims:
            # Note: this is clobbering of the claims above!  might be fine
            # for this test class, but be warned if you copy-paste somewhere.
            unsigned_jwt.update(extra_claims)

        headers = {}
        if self.signing_key_id:
            headers["kid"] = self.signing_key_id

        if remove_claims:
            for remove_claim in remove_claims:
                if remove_claim in unsigned_jwt:
                    unsigned_jwt.pop(remove_claim)

        signed_jwt = self.encode(body=unsigned_jwt, extra_headers=headers)
        return signed_jwt

    def generate_legacy_token(self, ttl, api_key):
        now = int(time.time())

        unsigned_jwt = {
            "program_id": 7,
            "token_type": "auth",
            "role_level": 9999,
            "organization_id": 99,
            "user_id": 999999,
            "plan_template_id": None,
            "membership_id": 123456,
            "organization_name": "Planet",
            "2fa": False,
            "exp": now + ttl,
            "user_name": "Test User",
            "email": "test.user@planet.com",
        }
        if api_key:
            unsigned_jwt["api_key"] = api_key

        signed_jwt = self.encode(body=unsigned_jwt, extra_headers=None)
        return signed_jwt


class TestTokenBuilder(UtilTokenBuilderBase):
    """
    Build tokens locally for testing purposes.
    """

    def __init__(self, signing_key_file, signing_key_algorithm="RS256", **kwargs):
        super().__init__(**kwargs)
        self.signing_key_file = signing_key_file
        self.signing_key = load_rsa_private_key(signing_key_file)
        self.signing_key_algorithm = signing_key_algorithm

    def encode(self, body, extra_headers):
        return jwt.encode(body, self.signing_key, algorithm=self.signing_key_algorithm, headers=extra_headers)

    @staticmethod
    def test_token_builder_factory(keypair_name):
        public_key_file = tdata_resource_file_path("keys/{}_pub_jwk.json".format(keypair_name))
        signing_key_file = tdata_resource_file_path("keys/{}_priv_nopassword.test_pem".format(keypair_name))

        with open(public_key_file, "r", encoding="UTF-8") as file_r:
            pubkey_data = json.load(file_r)

        signing_key_id = pubkey_data["kid"]
        signing_key_algorithm = pubkey_data["alg"]
        token_builder = TestTokenBuilder(
            issuer="test_token_issuer_for_" + keypair_name,
            audience="test_token_audience_for_" + keypair_name,
            signing_key_file=signing_key_file,
            signing_key_id=signing_key_id,
            signing_key_algorithm=signing_key_algorithm,
        )
        return pubkey_data, token_builder


class FakeTokenBuilder(UtilTokenBuilderBase):
    """
    Build fake tokens that have complete random garbage as the signature and lie about the signing algorithm.
    """

    def __init__(self, signing_key_algorithm="RS256", **kwargs):
        super().__init__(**kwargs)
        self.signing_key_algorithm = signing_key_algorithm

    @staticmethod
    def fake_token(header, body):
        return "{}.{}.{}".format(
            str(
                jwt.utils.base64url_encode(
                    bytes(
                        json.dumps(header),
                        "utf-8",
                    )
                ),
                encoding="utf-8",
            ),
            str(
                jwt.utils.base64url_encode(bytes(json.dumps(body), "utf-8")),
                encoding="utf-8",
            ),
            str(jwt.utils.base64url_encode(secrets.token_bytes(256)), encoding="utf-8"),
        )

    def encode(self, body, extra_headers):
        headers = {"alg": self.signing_key_algorithm}
        if extra_headers:
            headers.update(extra_headers)
        return self.fake_token(body=body, header=headers)


class StubOidcClientConfig(OidcAuthClientConfig):
    def __init__(
        self,
        stub_authority_ttl,
        stub_authority_access_token_audience,
        stub_authority_signing_key_file,
        stub_authority_pub_key_file,
        **kwargs,
    ):
        # A primary goal in this patching of the super class is to prevent
        # network reach out.  In many cases, we are shorting out OIDC discovery.
        if not "client_id" in kwargs:
            kwargs["client_id"] = "__stub_client_id__"

        kwargs["jwks_endpoint"] = "__stub_jwks_endpoint__"
        kwargs["issuer"] = kwargs["auth_server"]

        super().__init__(**kwargs)
        self.stub_authority_ttl = stub_authority_ttl
        self.stub_authority_access_token_audience = stub_authority_access_token_audience
        self.stub_authority_token_signing_key_file = stub_authority_signing_key_file
        self.stub_authority_pub_key_file = stub_authority_pub_key_file


class StubJwksApiClient(JwksApiClient):
    def __init__(self, jwks_uri, pubkey_file):
        super().__init__(jwks_uri=jwks_uri)
        self._keys = []
        with open(pubkey_file, "r", encoding="UTF-8") as file_r:
            pubkey_data = json.load(file_r)
            if "keys" in pubkey_data:
                for key in pubkey_data["keys"]:
                    self._keys.append(key)
            else:
                self._keys.append(pubkey_data)

    def _checked_fetch(self):
        return {"keys": self._keys}


class StubOidcAuthClient(OidcAuthClient):
    """
    A stub auth client that implements a token issuing authority that is entirely local.
    This can be used to mock out the OidcAuthClient in unit tests that are testing
    functionality that is built on top of the auth client.
    """

    def __init__(self, client_config: StubOidcClientConfig):
        super().__init__(client_config)
        self._mock_client_config = client_config
        self._refresh_state = {}

        # Monkey patch jwks to match our stub token issuer.
        # This looks so wrong. There has to be a better way to patch superclass variables?
        jwks_client = StubJwksApiClient(
            self._mock_client_config.jwks_endpoint(), pubkey_file=self._mock_client_config.stub_authority_pub_key_file
        )
        self._OidcAuthClient__jwks_client = jwks_client
        self.token_builder = TestTokenBuilder(
            issuer=self._mock_client_config.auth_server(),
            audience=self._mock_client_config.stub_authority_access_token_audience,
            signing_key_file=self._mock_client_config.stub_authority_token_signing_key_file,
            signing_key_id=jwks_client._keys[0]["kid"],
        )

    def _construct_oidc_credential(
        self,
        get_access_token: bool,
        get_id_token: bool,
        get_refresh_token: bool,
        username: str,
        requested_scopes: list,
        extra_claims,
        remove_claims,
    ):
        jwt_access_token = self.token_builder.construct_oidc_access_token_rfc8693(
            username=username,
            requested_scopes=requested_scopes,
            ttl=self._mock_client_config.stub_authority_ttl,
            extra_claims=extra_claims,
            remove_claims=remove_claims,
        )
        jwt_id_token = self.token_builder.construct_oidc_id_token(
            client_id=self._mock_client_config.client_id(),
            ttl=self._mock_client_config.stub_authority_ttl,
            extra_claims=extra_claims,
            remove_claims=remove_claims,
        )

        credential_data = {
            "token_type": "Bearer",
            "expires_in": self._mock_client_config.stub_authority_ttl,
            "scope": " ".join(requested_scopes),
        }
        if get_access_token:
            credential_data["access_token"] = jwt_access_token
        if get_id_token:
            credential_data["id_token"] = jwt_id_token
        if get_refresh_token:
            refresh_token = str(uuid.uuid4())
            credential_data["refresh_token"] = refresh_token
            self._refresh_state[refresh_token] = {
                "get_id_token": get_id_token,
                "get_access_token": get_access_token,
                "username": username,
                "extra_claims": extra_claims,
                "requested_scopes": requested_scopes,
            }

        credential = FileBackedOidcCredential(data=credential_data)
        return credential

    def _client_auth_enricher(self, raw_payload: EnricherPayloadType, audience: str) -> EnricherReturnType:
        # return raw_payload, None
        # Abstract in the base class. Not under test here.
        assert 0

    def _oidc_flow_login(
        self,
        allow_open_browser=False,
        allow_tty_prompt=False,
        requested_scopes=None,
        requested_audiences=None,
        extra=None,
        username=None,
        get_access_token=True,
        get_id_token=True,
        get_refresh_token=True,
        extra_claims=None,
        remove_claims=None,
        **kwargs,
    ) -> FileBackedOidcCredential:
        if not requested_scopes:
            requested_scopes = self._mock_client_config.scopes()
        return self._construct_oidc_credential(
            get_access_token=get_access_token,
            get_id_token=get_id_token,
            get_refresh_token=get_refresh_token,
            username=username,
            requested_scopes=requested_scopes,
            extra_claims=extra_claims,
            remove_claims=remove_claims,
        )

    def refresh(
        self, refresh_token: str, requested_scopes: List[str] = None, extra: Optional[dict] = None
    ) -> FileBackedOidcCredential:
        # FIXME?  This is a test stub implementation. We ignore requests
        #         to change scopes on refresh.
        if refresh_token:
            return self._construct_oidc_credential(
                get_refresh_token=True,
                get_access_token=self._refresh_state[refresh_token]["get_access_token"],
                get_id_token=self._refresh_state[refresh_token]["get_id_token"],
                username=self._refresh_state[refresh_token]["username"],
                requested_scopes=self._refresh_state[refresh_token]["requested_scopes"],
                extra_claims=self._refresh_state[refresh_token]["extra_claims"],
                remove_claims=None,
            )
        else:
            # raise AuthClientException("cannot refresh without refresh token")
            raise TokenApiException(message="cannot refresh without refresh token")

    def validate_access_token_remote(self, access_token: str) -> dict:
        # Stubbing out remote network calls.  Our "remote" validator will
        # do a local validation, then construct a matching RFC 7662 response
        local_validation = self.validate_access_token_local(
            access_token=access_token, required_audience=self._mock_client_config.stub_authority_access_token_audience
        )
        if not local_validation:  # Local validation should throw if invalid, but just in case...
            return {"active": False}

        if local_validation.get(_SCOPE_CLAIM_RFC8693):
            # RFC 8693 places scopes in a space delimited string.
            token_scopes = local_validation.get(_SCOPE_CLAIM_RFC8693).split()
        elif local_validation.get(_SCOPE_CLAIM_OKTA):
            # No split.  Okta places a list of strings in the token.
            token_scopes = local_validation.get(_SCOPE_CLAIM_OKTA)
        else:
            token_scopes = []

        return {
            "active": True,
            "iss": local_validation.get("iss"),
            "aud": local_validation.get("aud"),
            "sub": local_validation.get("sub"),
            "scope": " ".join(token_scopes),
            "exp": local_validation.get("exp"),
            "iat": local_validation.get("iat"),
            "jti": local_validation.get("jti"),
        }

    def default_request_authenticator(
        self, credential: Union[pathlib.Path, Credential]
    ) -> CredentialRequestAuthenticator:
        # Abstract in the base class. Not under test here.
        return ForbiddenRequestAuthenticator()

    def can_login_unattended(self) -> bool:
        return True


class MockStorageObjectNotFound(FileBackedJsonObjectException):
    pass


class MockObjectStorageProvider(ObjectStorageProvider):
    def __init__(self, initial_mock_storage):
        self._mock_storage = copy.deepcopy(initial_mock_storage)

    def _peek(self):
        return self._mock_storage

    def load_obj(self, key: ObjectStorageProvider_KeyType) -> dict:
        if key not in self._mock_storage:
            raise MockStorageObjectNotFound
        return self._mock_storage[key]

    def save_obj(self, key: ObjectStorageProvider_KeyType, data: dict) -> None:
        self._mock_storage[key] = data

    def obj_exists(self, key: ObjectStorageProvider_KeyType) -> bool:
        return key in self._mock_storage

    def mtime(self, key: ObjectStorageProvider_KeyType) -> float:
        # Fake storage provider.  We don't know when objects were last modified.
        return 0.0

    def obj_rename(self, src: ObjectStorageProvider_KeyType, dst: ObjectStorageProvider_KeyType) -> None:
        if self._mock_storage[src] is not None:
            self._mock_storage[dst] = self._mock_storage[src]
            del self._mock_storage[src]
        else:
            del self._mock_storage[dst]
